------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automatically check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff
-- Cheddar is developped by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
-----------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Package description :
--
-- You will find here every constants that can be tuned by users.
-- Change definitions contained in this package according to your requirements
-- (size of Cheddar XML files, memory available on your machine, ...)
--
---
------------------------------------------------------------------------------
------------------------------------------------------------------------------







package Config is


   pragma Elaborate_Body;




   ----------------------------------------------------------------------------------------
   --
   -- Constants defined under this comment can be tuned by users to match user 
   -- requirements in terms 
   -- of systems that Cheddar can analyse and in terms
   -- of available memory/CPU in the user's machine. 
   --
   -- WARNING : DO NOT FORGET TO CHANGE config/config.h
   --
   ----------------------------------------------------------------------------------------



   -- Maximum processors, tasks, messages, buffers and resources
   -- which can be defined in a cheddar XML file
   -- Increasing these constants implies increasing cheddar memory footprint
   --
   Max_Processors      : constant Natural := 30;  
   Max_Tasks           : constant Natural := 200;  
   Max_Messages        : constant Natural := 5;  
   Max_Buffers         : constant Natural := 30;  
   Max_Resources       : constant Natural := 30;  
   Max_Networks        : constant Natural := 5;  
   Max_Event_Analyzers : constant Natural := 5;  
   Max_Address_Space   : constant Natural := 30;  


   -- Define the largest scheduling interval time that the
   -- simulator can compute
   -- Increasing this constant implies increasing cheddar CPU usage
   --
   Max_Scheduling_Period : constant Natural := 15000;  


   -- Definition of priority range. We choose the most current value :
   -- 256 priority levels (from 0 to 255)
   -- Zero is the lowest priority level
   ---
   type Priority_Range is range 0 .. 255; 
   Low_Priority  : constant Priority_Range := Priority_Range'First;  
   High_Priority : constant Priority_Range := Priority_Range'Last;  


   -- Maximum accurrancy of big numbers packages
   -- Increase it if you need to run large XML Chedar project file
   -- Increasing this constant implies increasing cheddar memory footprint
   -- Increasing this constant implies increasing cheddar CPU usage
   --
   Max_Big_Number_Size : constant Natural := 200;  


   -- Define the language used in the graphical editor of Cheddar
   -- Default language is "english"
   --
   type Languages is 
         (Francais, 
          English); 
   Current_Language : Languages := English;  


   -- Constants above define maximum number of tasks that
   -- can use a shared resource or a buffer
   -- increasing this constant implies increasing cheddar memory footprint
   --
   Max_Tasks_For_A_Resource : constant Natural := 10;  
   Max_Tasks_For_A_Buffer   : constant Natural := 10;  


   -- Define maximum of variables that can be defined
   -- in parametric scheduler
   -- Increasing this constant implies increasing cheddar memory footprint
   --
   Max_Variables : constant Natural := 200;  


   -- Maximum number of "set" statement in a parametric
   -- scheduler
   -- Increasing this constant implies increasing cheddar memory footprint
   --
   Max_Statements : constant Natural := 100;  



   -- Maximum number of dependencies of a XML Cheddar project file.
   -- Dependencies can be a message/buffer used by two tasks or a precedencies
   -- constraint.
   -- Increasing this constant implies increasing cheddar memory footprint
   --
   Max_Tasks_Dependencies : constant Natural := 30;  


   -- Maximum number of time line displayed at simulation
   -- If the number of tasks/messages/buffers/resources in
   -- a project is greater that this constant, the user have
   -- to choose which object the simulator engine has to display
   --
   Max_Time_Line_To_Display : constant Natural := 30;  



   ----------------------------------------------------------------------------------------
   --
   -- Constants defined under this comment is less usefull for the
   -- user and should be unchanged by him.
   --
   --
   -- WARNING : DO NOT FORGET TO CHANGE config/config.h
   --
   ----------------------------------------------------------------------------------------
   -- Define maximum of user's defined task parameter
   --
   Max_User_Defined_Task_Parameters : constant Natural := 5;  


   -- Maximum of parameters during a framework call
   --
   Max_Framework_Parameters : constant Natural := 20;  


   -- Define offset constants : a task can contain at most 3 offsets
   --
   Max_Offsets : constant Natural := 5;  


   -- This constant define the size of "xi" of a discrete
   -- density probability
   --
   Max_Density_Size : constant Natural := 200;  


   -- Maximum size of the framework request table
   -- Increasing this constant implies increasing cheddar memory footprint
   --
   Max_Requests : constant Natural := 50;  



   -- Maximul size of a fixed string
   --
   Max_String_Size : constant Natural := 200;  



   -- This constant is used by the parametric simulation engine
   -- It defines size of tables which store simulation data for
   -- all elements of Cheddar project (except time units)
   --
   Max_Simulation : constant Natural := Natural'Max (Natural'Max (Max_Processors, Max_Tasks), Natural'Max (Natural'Max (Max_Messages, Max_Buffers), Natural'Max (Max_Resources, Max_Networks)));  


   -- Maximum number of time line which can de computed
   -- All these time line can not be displayed
   --
   Max_Time_Line : constant Natural := Max_Tasks + Max_Messages + Max_Resources + Max_Buffers;  



   -- Sum of all objects which can be stored in a XML project
   --
   Sum_Simulation : constant Natural := Max_Processors + Max_Tasks + Max_Messages + Max_Resources + Max_Networks;  





end Config;
