------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff, Jerome Legrand
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------





with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;
with Ada.Exceptions;
use Ada.Exceptions;


with Time_Units;
use  Time_Units;
use  Time_Units.A_Time_Unit;

with Unbounded_Strings;
use Unbounded_Strings;

with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;
with Tasks;
use Tasks;

with Resource_Set;
use Resource_Set;
use Resource_Set.Generic_Resource_Set;
with Resources;
use Resources;
use Resources.Task_Lists;

with Config;
use Config;
with Translate;
use Translate;

with Objects;
use Objects;

with Integer_Util;



package body Address_Space_Set is




   procedure Add_Address_Space (
         My_Address_Spaces : in out Address_Spaces_Set; 
         Name              : in     Unbounded_String;   
         Cpu_Name          : in     Unbounded_String;   
         Text_Memory_Size  : in     Integer;            
         Stack_Memory_Size : in     Integer;            
         Data_Memory_Size  : in     Integer;            
         Heap_Memory_Size  : in     Integer             ) is 

      Dummy : Address_Space_Ptr;  
   begin
      Add_Address_Space(My_Address_Spaces, Dummy, Name,  Cpu_Name,
         Text_Memory_Size, Stack_Memory_Size, Data_Memory_Size,
         Heap_Memory_Size);
   end Add_Address_Space;



   procedure Check_Address_Space (
         My_Address_Spaces : in     Address_Spaces_Set; 
         Name              : in     Unbounded_String;   
         Cpu_Name          : in     Unbounded_String;   
         Text_Memory_Size  : in     Integer;            
         Stack_Memory_Size : in     Integer;            
         Data_Memory_Size  : in     Integer;            
         Heap_Memory_Size  : in     Integer             ) is 

   begin


      if Cpu_Name=""
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Processor_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      end if;


      if Name=""
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Address_Space_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      end if;


      if not Is_A_Valid_Identifier(Name)
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Address_Space_Name(Current_Language) &
               Lb_Colon & Lb_Invalid_Identifier(Current_Language)));
      end if;


      if (Text_Memory_Size<0)
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Text_Memory_Size(Current_Language) &
               Lb_Must_Be(
                  Current_Language) & ">=0"));
      end if;


      if (Data_Memory_Size<0)
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Data_Memory_Size(Current_Language) &
               Lb_Must_Be(
                  Current_Language) & ">=0"));
      end if;



      if (Heap_Memory_Size<0)
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Heap_Memory_Size(Current_Language) &
               Lb_Must_Be(
                  Current_Language) & ">=0"));
      end if;



      if (Stack_Memory_Size<0)
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Address_Space(Current_Language) & " " & Name &
               " : " & Lb_Stack_Memory_Size(
                  Current_Language) &
               Lb_Must_Be(
                  Current_Language) & ">=0"));
      end if;



   end Check_Address_Space;



   procedure Add_Address_Space (
         My_Address_Spaces : in out Address_Spaces_Set; 
         A_Address_Space   : in out Address_Space_Ptr;  
         Name              : in     Unbounded_String;   
         Cpu_Name          : in     Unbounded_String;   
         Text_Memory_Size  : in     Integer;            
         Stack_Memory_Size : in     Integer;            
         Data_Memory_Size  : in     Integer;            
         Heap_Memory_Size  : in     Integer             ) is 


      My_Iterator : Address_Spaces_Iterator;  

   begin

      Check_Address_Space(My_Address_Spaces, Name,
         Cpu_Name,
         Text_Memory_Size,
         Stack_Memory_Size, Data_Memory_Size,
         Heap_Memory_Size);


      if(Get_Number_Of_Elements(My_Address_Spaces) > 0)
            then

         Reset_Iterator(My_Address_Spaces, My_Iterator);

         loop
            Current_Element(My_Address_Spaces, A_Address_Space,
               My_Iterator);
            if (Name = A_Address_Space.Name)
                  then
               Raise_Exception(Invalid_Parameter'
                  Identity,
                  To_String(Lb_Address_Space(Current_Language) & " " & Name &
                     " : " & Lb_Address_Space_Name(
                        Current_Language) &
                     Lb_Already_Defined(
                        Current_Language)));
            end if;

            exit when Is_Last_Element(
               My_Address_Spaces, My_Iterator);

            Next_Element(My_Address_Spaces,
               My_Iterator);
         end loop;
      end if;


      A_Address_Space:= new Address_Space;

      A_Address_Space.Cpu_Name:=Cpu_Name;
      A_Address_Space.Name:=Name;
      A_Address_Space.Text_Memory_Size:=Text_Memory_Size;
      A_Address_Space.Stack_Memory_Size:=Stack_Memory_Size;
      A_Address_Space.Data_Memory_Size:=Data_Memory_Size;
      A_Address_Space.Heap_Memory_Size:=Heap_Memory_Size;


      Add(My_Address_Spaces,A_Address_Space);


   exception
      when Generic_Address_Space_Set.Full_Set =>
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Can_Not_Define_More_Address_Spaces(
                  Current_Language)));

   end Add_Address_Space;





   function Search_Address_Space (
         My_Address_Spaces : in     Address_Spaces_Set; 
         Name              : in     Unbounded_String    ) 
     return Address_Space_Ptr is 


      My_Iterator : Address_Spaces_Iterator;  
      A_Addr      : Address_Space_Ptr;  
      Result      : Address_Space_Ptr;  

      Found : Boolean := False;  

   begin

      Reset_Iterator(My_Address_Spaces, My_Iterator);

      loop
         Current_Element(My_Address_Spaces,A_Addr,
            My_Iterator);

         if(A_Addr.Name = Name)
               then
            Found:=True;
            Result:=A_Addr;
         end if;

         exit when Is_Last_Element(My_Address_Spaces,
            My_Iterator);

         Next_Element(My_Address_Spaces, My_Iterator);

      end loop;

      if not Found
            then
         Raise_Exception(Address_Space_Not_Found'Identity,
            To_String(Lb_Address_Space_Name(Current_Language) & "=" & Name));
      end if;

      return Result;
   end Search_Address_Space;





   procedure Delete_Processor (
         My_Address_Spaces : in out Address_Spaces_Set; 
         A_Processor       : in     Unbounded_String    ) is 

      Tmp             : Address_Spaces_Set;  
      A_Address_Space : Address_Space_Ptr;  
      My_Iterator     : Address_Spaces_Iterator;  

   begin

      if(Get_Number_Of_Elements(My_Address_Spaces) >
            0)
            then

         Reset_Iterator(My_Address_Spaces, My_Iterator);
         loop
            Current_Element(My_Address_Spaces,
               A_Address_Space,
               My_Iterator);

            if(A_Address_Space.Cpu_Name = A_Processor)
                  then
               Add(Tmp, A_Address_Space);
            end if;

            exit when Is_Last_Element(
               My_Address_Spaces, My_Iterator);
            Next_Element(My_Address_Spaces,
               My_Iterator);
         end loop;

         if not Is_Empty(Tmp) then
            Reset_Iterator(My_Address_Spaces, My_Iterator);
            loop
               Current_Element(Tmp, A_Address_Space,
                  My_Iterator);


               Delete(My_Address_Spaces,
                  A_Address_Space);

               exit when Is_Last_Element(Tmp,
                  My_Iterator);
               Next_Element(Tmp, My_Iterator);
            end loop;

            Reset(Tmp,False);
         end if;
      end if;

   end Delete_Processor;




   --
   -- XML/AADL export sub-programs
   --



   function Export_Xml (
         My_Address_Spaces : in     Address_Spaces_Set ) 
     return Unbounded_String is 

      My_Iterator : Address_Spaces_Iterator;  
      A_Addr      : Address_Space_Ptr;  

      Result : Unbounded_String := Empty_String;  

   begin

      if not Is_Empty(My_Address_Spaces) then

         Result:= Result & To_Unbounded_String(
            "<address_spaces>") &
            Unbounded_Lf;
         Reset_Iterator(My_Address_Spaces, My_Iterator);

         loop
            Current_Element(My_Address_Spaces,A_Addr,
               My_Iterator);

            Result:= Result & To_Unbounded_String(
               Ascii.Ht &
               "<address_space>") & Unbounded_Lf;


            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<name>" & To_String(
                  A_Addr.Name) & "</name>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<cpu_name>" & To_String(
                  A_Addr.Cpu_Name) & "</cpu_name>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<text_memory_size>" &
               A_Addr.
               Text_Memory_Size'
               Img & "</text_memory_size>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<stack_memory_size>" &
               A_Addr.
               Stack_Memory_Size'
               Img & "</stack_memory_size>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<data_memory_size>" &
               A_Addr.
               Data_Memory_Size'
               Img & "</data_memory_size>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "<heap_memory_size>" &
               A_Addr.
               Heap_Memory_Size'
               Img & "</heap_memory_size>") &
               Unbounded_Lf;


            Result:= Result & To_Unbounded_String(
               Ascii.Ht &
               "</address_space>") & Unbounded_Lf;

            exit when Is_Last_Element(
               My_Address_Spaces, My_Iterator);

            Next_Element(My_Address_Spaces,
               My_Iterator);

         end loop;
         Result:= Result & To_Unbounded_String(
            "</address_spaces>") &
            Unbounded_Lf & Unbounded_Lf;
      end if;

      return Result;

   end Export_Xml;




   function Export_Aadl_Implementations (
         My_Address_Spaces : in     Address_Spaces_Set; 
         My_Tasks          : in     Tasks_Set;          
         My_Resources      : in     Resources_Set       ) 
     return Unbounded_String is 


      My_Iterator1    : Address_Spaces_Iterator;  
      A_Address_Space : Address_Space_Ptr;  

      My_Iterator2 : Tasks_Iterator;  
      A_Task       : Task_Ptr;  

      My_Iterator3 : Resources_Iterator;  
      A_Resource   : Resource_Ptr;  

      Result               : Unbounded_String := Empty_String;  
      Thread_Subcomponents : Unbounded_String := Empty_String;  
      Data_Subcomponents   : Unbounded_String := Empty_String;  
      First_Resource       : Boolean          := True;  

   begin


      if not Is_Empty(My_Address_Spaces) then


         Reset_Iterator(My_Address_Spaces, My_Iterator1);

         loop
            Current_Element(My_Address_Spaces,
               A_Address_Space,
               My_Iterator1);

            Result:= Result & To_Unbounded_String(
               "process " & To_String(
                  A_Address_Space.Name)) &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               "end " & To_String(
                  A_Address_Space.Name) & ";" ) &
               Unbounded_Lf;

            Result:=Result & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               "process implementation " & To_String(
                  A_Address_Space.
                  Name) & ".Impl") & Unbounded_Lf;


            Thread_Subcomponents :=  Export_Aadl_Declarations(My_Tasks,
               A_Address_Space.Name, 2);
            Data_Subcomponents :=  Export_Aadl_Declarations(My_Resources,
               A_Address_Space.Name, 2);
            if (Thread_Subcomponents /= Empty_String)
                  or (Data_Subcomponents /= Empty_String)
                  then
               Result:= Result & To_Unbounded_String(
                  Ascii.Ht &
                  "subcomponents") & Unbounded_Lf & Thread_Subcomponents &
                  Data_Subcomponents;
            end if;



            if Data_Subcomponents/=Empty_String then

               Reset_Iterator(My_Tasks, My_Iterator2);

               Result:= Result &
                  To_Unbounded_String( Ascii.Ht &
                  "connections") & Unbounded_Lf;

               loop
                  Current_Element(My_Tasks, A_Task,
                     My_Iterator2);

                  if A_Task.Address_Space_Name =
                        A_Address_Space.Name then
                     Result:= Result &
                        Export_Aadl_Connections(
                        My_Resources,
                        A_Task.Name,  2);
                  end if;

                  exit when Is_Last_Element(
                     My_Tasks, My_Iterator2);

                  Next_Element(My_Tasks,
                     My_Iterator2);

               end loop;
            end if;


            Result:= Result & To_Unbounded_String(
               Ascii.Ht &
               "properties") & Unbounded_Lf;

            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "Cheddar_Properties::Source_Global_Text_Size => " &
               A_Address_Space.Text_Memory_Size'
               Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "Cheddar_Properties::Source_Global_Data_Size => " &
               A_Address_Space.Data_Memory_Size'
               Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "Cheddar_Properties::Source_Global_Heap_Size => " &
               A_Address_Space.Heap_Memory_Size'
               Img & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(
               Ascii.Ht & Ascii.Ht &
               "Cheddar_Properties::Source_Global_Stack_Size => " &
               A_Address_Space.
               Stack_Memory_Size'Img & ";") &
               Unbounded_Lf;

            if Data_Subcomponents/=Empty_String then

               Result:= Result & To_Unbounded_String(  Ascii.Ht &
                  Ascii.Ht &
                  "Cheddar_Properties::Critical_Section => (")  &
                  Unbounded_Lf;

               Reset_Iterator(My_Resources, My_Iterator3);

               loop
                  Current_Element(My_Resources,
                     A_Resource,
                     My_Iterator3);

                  if A_Resource.Address_Space_Name =
                        A_Address_Space.Name then

                     if A_Resource.Task_Tab.Nb_Entries/=0 then
                        if not First_Resource then
                           Result:= Result & To_Unbounded_String(",") &
                              Unbounded_Lf;
                        end if;
                        First_Resource:=False;
                        for I in 0..A_Resource.Task_Tab.Nb_Entries-1 loop
                           Result:= Result & Ascii.Ht & Ascii.Ht &
                              Ascii.Ht
                              & """" & "instancied_" & To_String(
                              A_Resource.Name) & """" & "," & Unbounded_Lf;
                           Result:= Result &
                              To_Unbounded_String(
                              Ascii.Ht &
                              Ascii.Ht & Ascii.Ht
                              & """" & "instancied_" & To_String(
                                 A_Resource.Task_Tab.Entries(I).
                                 Item) & """,""" & To_String(
                                 Integer_Util.Format(
                                    A_Resource.Task_Tab.Entries(I).
                                    Data.Task_Begin)) &
                              """,""" & To_String(Integer_Util.Format(
                                    A_Resource.
                                    Task_Tab.Entries(I).Data.Task_End)) &
                              """");
                           if I/=A_Resource.Task_Tab.Nb_Entries-1 then
                              Result:= Result & To_Unbounded_String(  ",");
                              Result:=Result & Unbounded_Lf;
                           end if;
                        end loop;
                     end if;
                  end if;


                  exit when Is_Last_Element(
                     My_Resources, My_Iterator3);

                  Next_Element(My_Resources,
                     My_Iterator3);

               end loop;

               Result:= Result & To_Unbounded_String(Ascii.Ht &
                  Ascii.Ht &
                  Ascii.Ht & ");") & Unbounded_Lf;
            end if;

            Result:= Result & To_Unbounded_String(
               "end " & To_String(
                  A_Address_Space.Name) &
               ".Impl;" ) & Unbounded_Lf;

            Result:=Result & Unbounded_Lf;


            exit when Is_Last_Element(
               My_Address_Spaces, My_Iterator1);

            Next_Element(My_Address_Spaces,
               My_Iterator1);

         end loop;



      end if;

      return Result;

   end Export_Aadl_Implementations;



   function Export_Aadl_Declarations (
         My_Address_Spaces : in     Address_Spaces_Set; 
         Number_Of_Ht      : in     Natural             ) 
     return Unbounded_String is 

      My_Iterator     : Address_Spaces_Iterator;  
      A_Address_Space : Address_Space_Ptr;  

      Result : Unbounded_String := Empty_String;  


   begin

      if not Is_Empty(My_Address_Spaces) then


         Reset_Iterator(My_Address_Spaces, My_Iterator);

         loop
            Current_Element(My_Address_Spaces,
               A_Address_Space,
               My_Iterator);


            for I in 1..Number_Of_Ht loop
               Result:= Result & Ascii.Ht;
            end loop;

            Result:= Result & To_Unbounded_String(
               "instancied_" &
               To_String(A_Address_Space.
                  Name) &  " : process " & To_String(
                  A_Address_Space.Name) &
               ".Impl;") & Unbounded_Lf;

            exit when Is_Last_Element(
               My_Address_Spaces, My_Iterator);

            Next_Element(My_Address_Spaces,
               My_Iterator);

         end loop;



      end if;


      return Result;


   end Export_Aadl_Declarations;



   function Export_Aadl_Properties (
         My_Address_Spaces : in     Address_Spaces_Set; 
         Number_Of_Ht      : in     Natural             ) 
     return Unbounded_String is 


      My_Iterator     : Address_Spaces_Iterator;  
      A_Address_Space : Address_Space_Ptr;  

      Result : Unbounded_String := Empty_String;  


   begin

      if not Is_Empty(My_Address_Spaces) then


         Reset_Iterator(My_Address_Spaces, My_Iterator);

         loop
            Current_Element(My_Address_Spaces,
               A_Address_Space,
               My_Iterator);


            for I in 1..Number_Of_Ht loop
               Result:= Result & Ascii.Ht;
            end loop;

            Result:= Result & To_Unbounded_String(
               "Actual_Processor_Binding => reference instancied_" &
               To_String(A_Address_Space.
                  Cpu_Name) &
               " applies to instancied_" & To_String(
                  A_Address_Space.Name) & ";")  &
               Unbounded_Lf;

            exit when Is_Last_Element(
               My_Address_Spaces, My_Iterator);

            Next_Element(My_Address_Spaces,
               My_Iterator);

         end loop;



      end if;


      return Result;


   end Export_Aadl_Properties;






end Address_Space_Set;


