------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff, Jerome Legrand
-- Cheddar is developed by the LISYC Team
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_Io;
use Text_Io;

with Tasks;
use Tasks;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Unbounded_Strings;
use Unbounded_Strings;

with Time_Units;
use Time_Units;

with Double_Util;
use Double_Util;

with Config;
use Config;


package body Buffers is


   procedure Put (
         B : in     Buffer_Role ) is 
   begin
      Put(B.The_Role'Img & "/" & B.Size'Img & "/" & B.Time'Img);
   end Put;

   procedure Initialize (
         B :    out Buffer_Role ) is 
   begin
      B.The_Role:=No_Role;
      B.Size:=0;
      B.Time:=0;
   end Initialize;



   procedure Initialize (
         A_Buff : in out Buffer ) is 
   begin
      A_Buff.Object_Type:=Buffer_Object_Type;
      A_Buff.Cpu_Name:=Empty_String;
      A_Buff.Address_Space_Name:=Empty_String;
      A_Buff.Size:=0;
      Initialize(A_Buff.Roles);
   end Initialize;



   procedure Put_Name (
         A_Buff : in     Buffer_Ptr ) is 
   begin
      Put(To_String(A_Buff.Name));
   end Put_Name;

   procedure Put (
         A_Buff : in     Buffer_Ptr ) is 
   begin
      Put(A_Buff.All);
   end Put;

   procedure Put (
         My_Buff : in     Buffer ) is 
   begin
      Put("Buffer name : " & To_String(My_Buff.Name));
      Put(" ; Cpu : " & To_String(My_Buff.Cpu_Name));
      Put(" ; Address space : " & To_String(My_Buff.Address_Space_Name));
      Put(" ; Size : " & My_Buff.Size'Img);
      Put(" ; x : " & My_Buff.X'Img);
      Put(" ; y : " & My_Buff.Y'Img);
      New_Line;
      Put(My_Buff.Roles);
      New_Line;
   end Put;


   function Copy (
         My_Buffer : in     Buffer_Ptr ) 
     return Buffer_Ptr is 

      Ptr : Buffer_Ptr;  
   begin
      Ptr:=new Buffer;
      Ptr.All:=My_Buffer.All;

      return Ptr;
   end Copy;



   procedure Buffer_Flow_Control (
         My_Buff   : in     Buffer_Ptr; 
         My_Tasks  : in     Tasks_Set;  
         Flow_Cons : in out Double;     
         Flow_Prod : in out Double      ) is 

      A_Task : Task_Ptr;  

   begin
      --init
      Flow_Prod := 0.0;
      Flow_Cons := 0.0;

      -- compute data flow
      for I in  0 .. (My_Buff.Roles.Nb_Entries - 1) loop

         -- compute producers data flow
         if (My_Buff.Roles.Entries(I).Data.The_Role = Producer)
               then
            A_Task := Search_Task(My_Tasks,My_Buff.Roles.Entries(I).Item);

            case A_Task.Task_Type is
               when Periodic_Type =>
                  Flow_Prod := Flow_Prod +
                     Double(My_Buff.Roles.Entries(I).Data.Size) /
                     Double(Periodic_Task_Ptr(A_Task).Period);

               when Poisson_Type =>
                  Flow_Prod := Flow_Prod +
                     Double(My_Buff.Roles.Entries(I).Data.Size) /
                     Double(Poisson_Task_Ptr(A_Task).Period);
               when others =>
                  raise Flow_Constraint_Not_Respected;
            end case;


         end if;

         -- Compute consumers data flow
         --
         if (My_Buff.Roles.Entries(I).Data.The_Role = Consumer)
               then
            A_Task := Search_Task(My_Tasks,My_Buff.Roles.Entries(I).Item);

            case A_Task.Task_Type is
               when Periodic_Type =>
                  Flow_Cons := Flow_Cons +
                     Double(My_Buff.Roles.Entries(I).Data.Size) /
                     Double(Periodic_Task_Ptr(A_Task).Period);
               when Poisson_Type =>
                  Flow_Cons := Flow_Cons +
                     Double(My_Buff.Roles.Entries(I).Data.Size) /
                     Double(Poisson_Task_Ptr(A_Task).Period);
               when others =>
                  raise Flow_Constraint_Not_Respected;
            end case;
         end if;

      end loop;

      if (not Equal(Flow_Prod,Flow_Cons)) and  (Flow_Prod > Flow_Cons)
            then
         raise Flow_Constraint_Not_Respected;
      end if;
   end Buffer_Flow_Control;



   function Is_Cons_Prod_Harmonic (
         My_Buff  : in     Buffer_Ptr; 
         My_Tasks : in     Tasks_Set   ) 
     return Boolean is 

      A_Task1,  
      A_Task2 : Task_Ptr;  
      Result  : Boolean  := True;  
      Tmp_Mod : Natural;  

   begin

      for I in  0 .. (My_Buff.Roles.Nb_Entries - 1) loop
         A_Task1 := Search_Task(My_Tasks,My_Buff.Roles.Entries(I).Item);
         for J in  I .. (My_Buff.Roles.Nb_Entries - 1) loop
            A_Task2 := Search_Task(My_Tasks,My_Buff.Roles.Entries(J).Item);

            if A_Task1.Task_Type = Periodic_Type then

               if ( Periodic_Task_Ptr(A_Task1).Period >
                     Periodic_Task_Ptr(A_Task2).Period)
                     then
                  Tmp_Mod := Periodic_Task_Ptr(A_Task1).Period mod
                     Periodic_Task_Ptr(A_Task2).Period;
               else
                  Tmp_Mod := Periodic_Task_Ptr(A_Task2).Period mod
                     Periodic_Task_Ptr(A_Task1).Period;
               end if;

            else
               if A_Task1.Task_Type = Poisson_Type then

                  if ( Poisson_Task_Ptr(A_Task1).Period >
                        Poisson_Task_Ptr(A_Task2).Period)
                        then
                     Tmp_Mod := Poisson_Task_Ptr(A_Task1).Period mod
                        Poisson_Task_Ptr(A_Task2).Period;
                  else
                     Tmp_Mod := Poisson_Task_Ptr(A_Task2).Period mod
                        Poisson_Task_Ptr(A_Task1).Period;
                  end if;
               end if;
            end if;

            if (Tmp_Mod /= 0 )
                  then
               Result:=False;
            end if;
         end loop;
      end loop;
      return Result;
   end Is_Cons_Prod_Harmonic;








end Buffers;





