------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff, Jerome Legrand
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------


with Text_Io;
use Text_Io;


with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;


with Config;

with Scheduler;
use Scheduler;

with Processors;
use Processors;

with Sets;



package Processor_Set is




   ----------------------------------------------------------
   -- Definition of a set
   ----------------------------------------------------------

   package Generic_Processor_Set is new Sets (
      Max_Element => Config.Max_Processors, 
      Element     => Processor_Ptr,         
      Free        => Free,                  
      Copy        => Copy,                  
      Put         => Put);


   use Generic_Processor_Set;

   type Processors_Set is new Generic_Processor_Set.Set with private; 


   subtype Processors_Range is Generic_Processor_Set.Element_Range;
   subtype Processors_Iterator is   Generic_Processor_Set.Iterator;




   ----------------------------------------------------------
   -- Exceptions defined in the set package
   ----------------------------------------------------------

   -- Raised when a given processor is not in a processor set
   --
   Processor_Not_Found : exception;  

   -- Raised when parameters provided to Add_processor are wrong
   --
   Invalid_Parameter : exception;  

   -- Raised when the scheduler code of a parametric scheduler
   -- does not have a right syntax
   --
   Scheduler_Syntax_Error : exception;  




   ----------------------------------------------------------
   -- I/O operations
   ----------------------------------------------------------
   function Export_Aadl_Implementations (
         My_Processors : in     Processors_Set ) 
     return Unbounded_String; 

   function Export_Aadl_Declarations (
         My_Processors : in     Processors_Set; 
         Number_Of_Ht  : in     Natural         ) 
     return Unbounded_String; 

   function Export_Xml (
         My_Processors : in     Processors_Set ) 
     return Unbounded_String; 



   ----------------------------------------------------------
   -- Procedure to proceed modification on a set
   ----------------------------------------------------------


   procedure Check_Processor (
         My_Processors : in     Processors_Set;   
         Name          : in     Unbounded_String; 
         Is_Preemptive : in     Preemptive_Type;  
         Quantum       : in     Integer;          
         File_Name     : in     Unbounded_String; 
         A_Scheduler   : in     Scheduler_Type;   
         A_Network     : in     Unbounded_String  ); 

   procedure Add_Processor (
         My_Processors : in out Processors_Set;   
         A_Processor   : in out Processor_Ptr;    
         Name          : in     Unbounded_String; 
         Is_Preemptive : in     Preemptive_Type;  
         Quantum       : in     Integer;          
         File_Name     : in     Unbounded_String; 
         A_Scheduler   : in     Scheduler_Type;   
         A_Network     : in     Unbounded_String  ); 

   procedure Add_Processor (
         My_Processors : in out Processors_Set;   
         Name          : in     Unbounded_String; 
         Is_Preemptive : in     Preemptive_Type;  
         Quantum       : in     Integer;          
         File_Name     : in     Unbounded_String; 
         A_Scheduler   : in     Scheduler_Type;   
         A_Network     : in     Unbounded_String  ); 


   ------------------------------------------------------
   -- Read information from a set
   ------------------------------------------------------


   -- Get the task pointer of the processor named "Name"
   --
   function Search_Processor (
         My_Processors : in     Processors_Set;  
         Name          : in     Unbounded_String ) 
     return Processor_Ptr; 



   ----------------------------------------------------------
   -- List of parameters the user can set of get directly
   -- throught the set
   ----------------------------------------------------------



   type Processor_Parameters is 
         (Scheduler); 


   function Get (
         My_Processors  : in     Processors_Set;      
         Processor_Name : in     Unbounded_String;    
         Param_Name     : in     Processor_Parameters ) 
     return Scheduler_Ptr; 

   procedure Set (
         My_Processors  : in out Processors_Set;       
         Processor_Name : in     Unbounded_String;     
         Param_Name     : in     Processor_Parameters; 
         Param_Value    : in     Scheduler_Ptr         ); 




private

   type Processors_Set is new Generic_Processor_Set.Set with null record; 




end Processor_Set;
