------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automatically check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff, Jerome Legrand
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Resources;
use Resources;

with Time_Units;
use  Time_Units;
use  Time_Units.A_Time_Unit;

with Unbounded_Strings;
use Unbounded_Strings;


with Ada.Exceptions;
use Ada.Exceptions;


with Config;
use Config;
with Translate;
use Translate;
with Objects;
use Objects;



package body Resource_Set is


   procedure Task_Begin_End (
         Task_Name      : in     Unbounded_String; 
         Resource_Name  : in     Unbounded_String; 
         A_Resource_Set : in     Resources_Set;    
         Starts_Ends    : in out Start_End_Table;  
         Elements       : in out Task_List_Range   ) is 

      My_Resource_Iterator : Resources_Iterator;  
      A_Resource           : Resource_Ptr;  
      Index                : Task_List_Range    := 0;  

      -- looking for a resource and a task the end and the begin of this task on this resource
      -- Element is the number of element, must start to 0
      --
   begin
      Reset_Iterator(A_Resource_Set, My_Resource_Iterator);
      loop
         Current_Element(A_Resource_Set, A_Resource, My_Resource_Iterator);
         if A_Resource.Name = Resource_Name then

            for I in 0..A_Resource.Task_Tab.Nb_Entries-1 loop
               if (A_Resource.Task_Tab.Entries(Index).Item= Task_Name) then
                  Starts_Ends(Elements+1).Task_Begin:= A_Resource.
                     Task_Tab.Entries(Index).Data.Task_Begin;
                  Starts_Ends(Elements+1).Task_End:= A_Resource.Task_Tab.
                     Entries(Index).Data.Task_End;
                  Elements := Elements + 1;
               end if;
               Index := Index + 1;
            end loop;
            exit;
         end if;
         exit when Is_Last_Element(A_Resource_Set, My_Resource_Iterator);
         Next_Element(A_Resource_Set, My_Resource_Iterator);
      end loop;

   end Task_Begin_End;




   procedure Update_Resource (
         My_Resources       : in out Resources_Set;    
         Name               : in     Unbounded_String; 
         State              : in     Integer;          
         Cpu_Name           : in     Unbounded_String; 
         Address_Space_Name : in     Unbounded_String; 
         Protocol           : in     Resources_Type;   
         Affected_Tasks     : in     Task_List         ) is 

      The_Resource : Resource_Ptr;  

   begin

      The_Resource := Search_Resource(My_Resources, Name);

      Check_Resource(My_Resources, Name, State, Cpu_Name,
         Address_Space_Name,
         Protocol);
      Delete(My_Resources, The_Resource);
      Add_Resource(My_Resources, Name, State, Cpu_Name, Address_Space_Name,
         Protocol, Affected_Tasks);

   end Update_Resource;


   procedure Add_Resource (
         My_Resources       : in out Resources_Set;    
         Name               : in     Unbounded_String; 
         State              : in     Integer;          
         Cpu_Name           : in     Unbounded_String; 
         Address_Space_Name : in     Unbounded_String; 
         Protocol           : in     Resources_Type;   
         Affected_Tasks     : in     Task_List         ) is 


      A_Resource_Ptr : Resource_Ptr;  

   begin

      Add_Resource(My_Resources, A_Resource_Ptr, Name, State, Cpu_Name,
         Address_Space_Name, Protocol,
         Affected_Tasks);

   end Add_Resource;



   procedure Check_Resource_Critical_Section (
         My_Resources  : in     Resources_Set;    
         Resource_Name : in     Unbounded_String; 
         Task_Name     : in     Unbounded_String; 
         Share_Begin   : in     Integer;          
         Share_End     : in     Integer           ) is 


   begin

      if (Share_Begin<=0)
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(
               Resource_Name & " ; " & Task_Name & " ; " &
               Lb_Task_Begin(Current_Language) & Lb_Must_Be(
                  Current_Language)
               & To_Unbounded_String(" >0 ")
               ));
      end if;

      if (Share_End<=0)
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(
               Resource_Name & " ; " & Task_Name & " ; " &
               Lb_Task_End(Current_Language) & Lb_Must_Be(
                  Current_Language)
               & To_Unbounded_String(" >0 ")
               ));
      end if;

      if Share_Begin > Share_End
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(
               Resource_Name & " ; " & Task_Name & " ; " &
               Lb_Task_Begin(Current_Language) & Lb_Must_Be(
                  Current_Language)
               & To_Unbounded_String(" <= ")
               & Lb_Task_End(Current_Language) ));
      end if;


   end Check_Resource_Critical_Section;


   procedure Check_Resource (
         My_Resources       : in     Resources_Set;    
         Name               : in     Unbounded_String; 
         State              : in     Integer;          
         Cpu_Name           : in     Unbounded_String; 
         Address_Space_Name : in     Unbounded_String; 
         Protocol           : in     Resources_Type    ) is 

   begin

      if Name=""
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Resource_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      end if;

      if not Is_A_Valid_Identifier(Name)
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Resource(Current_Language) & " " & Name &
               " : " & Lb_Resource_Name(Current_Language) &
               Lb_Colon & Lb_Invalid_Identifier(Current_Language)));
      end if;

      if Cpu_Name=""
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Resource(Current_Language) & " " & Name &
               " : " & Lb_Processor_Name(Current_Language) &
               Lb_Mandatory(Current_Language)));
      end if;


      if Address_Space_Name=""
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Resource(Current_Language) & " " & Name &
               " : " & Lb_Address_Space_Name(
                  Current_Language) &
               Lb_Mandatory(Current_Language)));
      end if;


      if (Protocol /= No_Protocol)
            and (Protocol /= Priority_Ceiling_Protocol)
            and (Protocol /= Priority_Inheritance_Protocol)
            then
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Resource(Current_Language) & " " & Name &
               " : " & Lb_Invalid_Protocol(Current_Language)));
      end if;


   end Check_Resource;



   procedure Add_Resource (
         My_Resources       : in out Resources_Set;    
         A_Resource         : in out Resource_Ptr;     
         Name               : in     Unbounded_String; 
         State              : in     Integer;          
         Cpu_Name           : in     Unbounded_String; 
         Address_Space_Name : in     Unbounded_String; 
         Protocol           : in     Resources_Type;   
         Affected_Tasks     : in     Task_List         ) is 


      New_Protocol     : Resources_Type;  
      New_Pcp_Resource : Pcp_Resource_Ptr;  
      New_Pip_Resource : Pip_Resource_Ptr;  
      New_Np_Resource  : Np_Resource_Ptr;  

      My_Iterator : Iterator;  

   begin


      Check_Resource(My_Resources, Name, State, Cpu_Name,
         Address_Space_Name, Protocol);



      -- Resource already exist ?
      --
      if(Get_Number_Of_Elements(My_Resources) > 0) then

         Reset_Iterator(My_Resources, My_Iterator);

         loop
            Current_Element(My_Resources, A_Resource, My_Iterator);

            if (Name = A_Resource.Name) then
               Raise_Exception(Invalid_Parameter'
                  Identity,
                  To_String(Lb_Resource(Current_Language) & " " & Name &
                     " : " & Lb_Resource_Name(
                        Current_Language) &
                     Lb_Already_Defined(
                        Current_Language)));
            end if;

            exit when Is_Last_Element(My_Resources, My_Iterator);

            Next_Element(My_Resources, My_Iterator);

         end loop;

      end if;


      if (Protocol = Priority_Ceiling_Protocol)
            then
         New_Pcp_Resource := new Pcp_Resource;
         New_Protocol := Priority_Ceiling_Protocol;
         A_Resource:=Resource_Ptr(New_Pcp_Resource);

      else
         if (Protocol = Priority_Inheritance_Protocol)
               then
            New_Pip_Resource := new Pip_Resource;
            New_Protocol := Priority_Inheritance_Protocol;
            A_Resource:=Resource_Ptr(New_Pip_Resource);

         else
            New_Np_Resource := new Np_Resource;
            New_Protocol := No_Protocol;
            A_Resource:=Resource_Ptr(New_Np_Resource);
         end if;
      end if;

      A_Resource.Name:=Name;
      A_Resource.State:=State;
      A_Resource.Cpu_Name:=Cpu_Name;
      A_Resource.Address_Space_Name:=Address_Space_Name;
      A_Resource.Protocol := New_Protocol;

      A_Resource.Task_Tab:=Affected_Tasks;


      Add(My_Resources,A_Resource);


   exception
      when Full_Set =>
         Raise_Exception(Invalid_Parameter'Identity,
            To_String(Lb_Can_Not_Define_More_Resources(
                  Current_Language)));

   end Add_Resource;




   function Search_Resource (
         My_Resources : in     Resources_Set;   
         Name         : in     Unbounded_String ) 
     return Resource_Ptr is 

      My_Iterator : Iterator;  
      A_Resource  : Resource_Ptr;  
      Result      : Resource_Ptr;  

      Found : Boolean := False;  

   begin

      Reset_Iterator(My_Resources, My_Iterator);

      loop
         Current_Element(My_Resources,A_Resource,My_Iterator);

         if(A_Resource.Name = Name) then
            Found:=True;
            Result:=A_Resource;
         end if;

         exit when Is_Last_Element(My_Resources, My_Iterator);

         Next_Element(My_Resources, My_Iterator);

      end loop;

      if not Found then
         Raise_Exception(Resource_Not_Found'Identity,
            To_String(Lb_Resource_Name(Current_Language) & "=" & Name));
      end if;

      return Result;

   end Search_Resource;

   procedure Delete_Task (
         My_Resources : in out Resources_Set;   
         A_Task       : in     Unbounded_String ) is 

      A_Resource  : Resource_Ptr;  
      My_Iterator : Resources_Iterator;  
      Index       : Task_List_Range    := 1;  

   begin

      if(Get_Number_Of_Elements(My_Resources) > 0) then

         Reset_Iterator(My_Resources, My_Iterator);
         loop
            Current_Element(My_Resources, A_Resource, My_Iterator);

            -- Looking for a_task in the task_list and delete it
            --
            Index:=0;
            while (Index<A_Resource.Task_Tab.Nb_Entries)
                  loop
               if (A_Resource.Task_Tab.Entries(Index).Item =
                     A_Task)
                     then
                  A_Resource.Task_Tab.Entries(Index):=
                     A_Resource.Task_Tab.
                     Entries(A_Resource.Task_Tab.Nb_Entries-1);

                  A_Resource.Task_Tab.Nb_Entries:=
                     A_Resource.Task_Tab.Nb_Entries-1;
               end if;
               Index:=Index+1;
            end loop;


            exit when Is_Last_Element(My_Resources, My_Iterator);
            Next_Element(My_Resources, My_Iterator);
         end loop;
      end if;

   end Delete_Task;



   procedure Delete_Address_Space (
         My_Resources : in out Resources_Set;   
         A_Addr       : in     Unbounded_String ) is 

      Tmp         : Resources_Set;  
      A_Resource  : Resource_Ptr;  
      My_Iterator : Resources_Iterator;  

   begin
      if(Get_Number_Of_Elements(My_Resources) > 0) then

         Reset_Iterator(My_Resources, My_Iterator);
         loop
            Current_Element(My_Resources, A_Resource, My_Iterator);

            if(A_Resource.Address_Space_Name = A_Addr)
                  then
               Add(Tmp, A_Resource);
            end if;

            exit when Is_Last_Element(My_Resources, My_Iterator);
            Next_Element(My_Resources, My_Iterator);
         end loop;

         if not Is_Empty(Tmp) then
            Reset_Iterator(Tmp, My_Iterator);
            loop
               Current_Element(Tmp, A_Resource, My_Iterator);

               Delete(My_Resources,A_Resource);

               exit when Is_Last_Element(Tmp, My_Iterator);
               Next_Element(Tmp, My_Iterator);
            end loop;
         end if;

      end if;
   end Delete_Address_Space;






   procedure Delete_Processor (
         My_Resources : in out Resources_Set;   
         A_Processor  : in     Unbounded_String ) is 

      Tmp         : Resources_Set;  
      A_Resource  : Resource_Ptr;  
      My_Iterator : Resources_Iterator;  

   begin

      if(Get_Number_Of_Elements(My_Resources) > 0)
            then

         Reset_Iterator(My_Resources, My_Iterator);
         loop
            Current_Element(My_Resources, A_Resource,
               My_Iterator);

            if(A_Resource.Cpu_Name = A_Processor)
                  then
               Add(Tmp, A_Resource);
            end if;

            exit when Is_Last_Element(My_Resources, My_Iterator);
            Next_Element(My_Resources, My_Iterator);
         end loop;

         if not Is_Empty(Tmp) then
            Reset_Iterator(Tmp, My_Iterator);
            loop
               Current_Element(Tmp, A_Resource, My_Iterator);


               Delete(My_Resources,A_Resource);

               exit when Is_Last_Element(Tmp, My_Iterator);
               Next_Element(Tmp, My_Iterator);
            end loop;

            Reset(Tmp,False);
         end if;
      end if;

   end Delete_Processor;




   procedure Same_Protocol_Control (
         My_Resources : in     Resources_Set;   
         A_Task       : in     Unbounded_String ) is 

      A_Protocol : Resources_Type;  
      Is_First   : Boolean            := True;  
      Iterator1  : Resources_Iterator;  
      Resource1  : Resource_Ptr;  

   begin

      Reset_Iterator(My_Resources, Iterator1);

      loop
         Current_Element(My_Resources, Resource1, Iterator1);

         -- Looking for task user in the task_list index_table
         --
         for I in 0..Resource1.Task_Tab.Nb_Entries-1 loop
            if A_Task = Resource1.Task_Tab.Entries(I).Item
                  then
               if  Is_First
                     then
                  Is_First:=False;
                  A_Protocol:=Resource1.Protocol;
               else
                  if A_Protocol /= Resource1.Protocol
                        then
                     raise Can_Not_Used_Different_Protocol;
                  end if;
               end if;
            end if;
         end loop;

         exit when Is_Last_Element(My_Resources, Iterator1);
         Next_Element(My_Resources, Iterator1);

      end loop;

   end Same_Protocol_Control;



   function Export_Xml (
         My_Resources : in     Resources_Set ) 
     return Unbounded_String is 

      My_Iterator : Resources_Iterator;  
      A_Resource  : Resource_Ptr;  

      Result : Unbounded_String := Empty_String;  

   begin

      if not Is_Empty(My_Resources) then

         Result:= Result & To_Unbounded_String( "<resources>") &
            Unbounded_Lf;

         Reset_Iterator(My_Resources, My_Iterator);

         loop
            Current_Element(My_Resources,A_Resource,My_Iterator);

            Result:= Result & To_Unbounded_String( Ascii.Ht &
               "<resource>") &
               Unbounded_Lf;

            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<cpu_name>" & To_String(
                  A_Resource.Cpu_Name) & "</cpu_name>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<address_space_name>" &
               To_String(
                  A_Resource.Address_Space_Name) &
               "</address_space_name>") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<name>" & To_String(
                  A_Resource.Name) & "</name>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<state>" & A_Resource.
               State'Img & "</state>") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<protocol>" & A_Resource.
               Protocol'Img & "</protocol>") & Unbounded_Lf;


            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "<resource_used_by>") & Unbounded_Lf;
            for I in 0..A_Resource.Task_Tab.Nb_Entries-1 loop
               Result:= Result & To_Unbounded_String( Ascii.Ht &
                  Ascii.Ht &
                  Ascii.Ht & "<resource_user>");
               Result:= Result & To_Unbounded_String( " " &
                  To_String(
                     A_Resource.Task_Tab.Entries(I).
                     Item) & " ");
               Result:= Result & To_Unbounded_String(
                  A_Resource.Task_Tab.Entries(I).Data.Task_Begin'
                  Img & " " );
               Result:= Result & To_Unbounded_String(
                  A_Resource.Task_Tab.Entries(I).Data.Task_End'
                  Img &
                  " " );
               Result:= Result & To_Unbounded_String(
                  "</resource_user>") &
                  Unbounded_Lf;
            end loop;
            Result:= Result & To_Unbounded_String( Ascii.Ht &
               Ascii.Ht &
               "</resource_used_by>") & Unbounded_Lf;


            Result:= Result & To_Unbounded_String( Ascii.Ht &
               "</resource>") & Unbounded_Lf;


            exit when Is_Last_Element(My_Resources, My_Iterator);

            Next_Element(My_Resources, My_Iterator);

         end loop;
         Result:= Result & To_Unbounded_String( "</resources>") &
            Unbounded_Lf;
      end if;

      return Result;

   end Export_Xml;



   function Export_Aadl_Implementations (
         My_Resources : in     Resources_Set ) 
     return Unbounded_String is 

      My_Iterator : Resources_Iterator;  
      A_Resource  : Resource_Ptr;  

      Result : Unbounded_String := Empty_String;  

   begin

      if not Is_Empty(My_Resources) then

         Reset_Iterator(My_Resources, My_Iterator);

         loop
            Current_Element(My_Resources,A_Resource,My_Iterator);

            Result:= Result & To_Unbounded_String( "data " &
               To_String(
                  A_Resource.Name)) & Unbounded_Lf;
            Result:= Result & To_Unbounded_String( "end " &
               To_String(
                  A_Resource.Name) & ";" ) & Unbounded_Lf &
               Unbounded_Lf;


            Result:= Result & To_Unbounded_String(
               "data implementation " &
               To_String(
                  A_Resource.
                  Name) & ".Impl" ) & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               "properties ") &
               Unbounded_Lf;
            Result:= Result & To_Unbounded_String(Ascii.Ht &
               Ascii.Ht &
               "Cheddar_Properties::Data_Concurrency_State => " &
               A_Resource.State'Img  & ";") & Unbounded_Lf;
            Result:= Result & To_Unbounded_String(  Ascii.Ht &
               Ascii.Ht &
               "Concurrency_Control_Protocol => " &
               A_Resource.Protocol'
               Img & ";") & Unbounded_Lf;

            Result:= Result & To_Unbounded_String(  "end " &
               To_String(
                  A_Resource.Name) & ".Impl;" ) & Unbounded_Lf &
               Unbounded_Lf;


            exit when Is_Last_Element(My_Resources, My_Iterator);

            Next_Element(My_Resources, My_Iterator);

         end loop;

      end if;

      return Result;

   end Export_Aadl_Implementations;



   function Export_Aadl_Declarations (
         My_Resources       : in     Resources_Set;    
         Address_Space_Name :        Unbounded_String; 
         Number_Of_Ht       : in     Natural           ) 
     return Unbounded_String is 


      My_Iterator : Resources_Iterator;  
      A_Resource  : Resource_Ptr;  

      Result : Unbounded_String := Empty_String;  

   begin


      if not Is_Empty(My_Resources) then


         Reset_Iterator(My_Resources, My_Iterator);

         loop
            Current_Element(My_Resources,A_Resource,My_Iterator);

            if A_Resource.Address_Space_Name = Address_Space_Name then
               for I in 1..Number_Of_Ht loop
                  Result:= Result & Ascii.Ht;
               end loop;
               Result:= Result & To_Unbounded_String(
                  "instancied_" &
                  To_String(A_Resource.Name) &
                  " : data " & To_String(A_Resource.Name) &
                  ".Impl;") &
                  Unbounded_Lf;
            end if;


            exit when Is_Last_Element(My_Resources, My_Iterator);

            Next_Element(My_Resources, My_Iterator);

         end loop;
      end if;

      return Result;

   end Export_Aadl_Declarations;


   function Export_Aadl_Connections (
         My_Resources : in     Resources_Set;    
         A_Task_Name  : in     Unbounded_String; 
         Number_Of_Ht : in     Natural           ) 
     return Unbounded_String is 

      My_Iterator : Resources_Iterator;  
      A_Resource  : Resource_Ptr;  

      Result : Unbounded_String := Empty_String;  

   begin
      if not Is_Empty(My_Resources) then


         Reset_Iterator(My_Resources, My_Iterator);

         loop
            Current_Element(My_Resources, A_Resource, My_Iterator);

            if A_Resource.Task_Tab.Nb_Entries/=0 then
               for I in 0..A_Resource.Task_Tab.Nb_Entries-1 loop
                  if  A_Resource.Task_Tab.Entries(I).Item =
                        A_Task_Name then
                     for J in 1..Number_Of_Ht loop
                        Result:= Result & Ascii.Ht;
                     end loop;
                     Result:= Result & To_Unbounded_String(
                        "data access " &
                        To_String("instancied_" & A_Task_Name &
                           "." & A_Resource.Name & "_features" ) &
                        " -> instancied_" & To_String(
                           A_Resource.Name) &
                        ";") &
                        Unbounded_Lf;
                  end if;
               end loop;
            end if;

            exit when Is_Last_Element(My_Resources, My_Iterator);

            Next_Element(My_Resources, My_Iterator);

         end loop;
      end if;

      return Result;

   end   Export_Aadl_Connections;




end Resource_Set;








