------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff, Jerome Legrand
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_Io;
use Text_Io;


with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Config;
use Config;

with Resources;
use Resources;
use Resources.Task_Lists;


with Sets;


package Resource_Set is




   ----------------------------------------------------------
   -- Definition of a shared resource set
   ----------------------------------------------------------



   package Generic_Resource_Set is new Sets (
      Max_Element => Config.Max_Resources, 
      Element     => Resource_Ptr,         
      Free        => Free,                 
      Copy        => Copy,                 
      Put         => Put);

   use Generic_Resource_Set;

   type Resources_Set is new Generic_Resource_Set.Set with private; 


   subtype Resources_Range is Generic_Resource_Set.Element_Range;
   subtype Resources_Iterator is Generic_Resource_Set.Iterator;



   ----------------------------------------------------------
   -- Exceptions defined in the resource set package
   ----------------------------------------------------------

   -- Try to get/set a unkwon parameter
   --
   Invalid_Parameter : exception;  

   -- The the resource name does ot exist in the set
   --
   Resource_Not_Found : exception;  

   -- all the resource  accessed by a task should have
   -- the same protocol type
   --
   Can_Not_Used_Different_Protocol : exception;  



   ----------------------------------------------------------
   -- I/O operations
   ----------------------------------------------------------
   function Export_Xml (
         My_Resources : in     Resources_Set ) 
     return Unbounded_String; 

   function Export_Aadl_Implementations (
         My_Resources : in     Resources_Set ) 
     return Unbounded_String; 

   function Export_Aadl_Declarations (
         My_Resources       : in     Resources_Set;    
         Address_Space_Name : in     Unbounded_String; 
         Number_Of_Ht       : in     Natural           ) 
     return Unbounded_String; 

   function Export_Aadl_Connections (
         My_Resources : in     Resources_Set;    
         A_Task_Name  : in     Unbounded_String; 
         Number_Of_Ht : in     Natural           ) 
     return Unbounded_String; 




   ----------------------------------------------------------
   -- Procedure to perform modification on a set
   ----------------------------------------------------------



   procedure Task_Begin_End (
         Task_Name      : in     Unbounded_String; 
         Resource_Name  : in     Unbounded_String; 
         A_Resource_Set : in     Resources_Set;    
         Starts_Ends    : in out Start_End_Table;  
         Elements       : in out Task_List_Range   ); 


   procedure Update_Resource (
         My_Resources       : in out Resources_Set;    
         Name               : in     Unbounded_String; 
         State              : in     Integer;          
         Cpu_Name           : in     Unbounded_String; 
         Address_Space_Name : in     Unbounded_String; 
         Protocol           : in     Resources_Type;   
         Affected_Tasks     : in     Task_List         ); 


   procedure Check_Resource (
         My_Resources       : in     Resources_Set;    
         Name               : in     Unbounded_String; 
         State              : in     Integer;          
         Cpu_Name           : in     Unbounded_String; 
         Address_Space_Name : in     Unbounded_String; 
         Protocol           : in     Resources_Type    ); 


   procedure Check_Resource_Critical_Section (
         My_Resources  : in     Resources_Set;    
         Resource_Name : in     Unbounded_String; 
         Task_Name     : in     Unbounded_String; 
         Share_Begin   : in     Integer;          
         Share_End     : in     Integer           ); 


   procedure Add_Resource (
         My_Resources       : in out Resources_Set;    
         Name               : in     Unbounded_String; 
         State              : in     Integer;          
         Cpu_Name           : in     Unbounded_String; 
         Address_Space_Name : in     Unbounded_String; 
         Protocol           : in     Resources_Type;   
         Affected_Tasks     : in     Task_List         ); 


   procedure Add_Resource (
         My_Resources       : in out Resources_Set;    
         A_Resource         : in out Resource_Ptr;     
         Name               : in     Unbounded_String; 
         State              : in     Integer;          
         Cpu_Name           : in     Unbounded_String; 
         Address_Space_Name : in     Unbounded_String; 
         Protocol           : in     Resources_Type;   
         Affected_Tasks     : in     Task_List         ); 


   procedure Delete_Task (
         My_Resources : in out Resources_Set;   
         A_Task       : in     Unbounded_String ); 


   procedure Delete_Address_Space (
         My_Resources : in out Resources_Set;   
         A_Addr       : in     Unbounded_String ); 


   procedure Delete_Processor (
         My_Resources : in out Resources_Set;   
         A_Processor  : in     Unbounded_String ); 




   ------------------------------------------------------
   -- Read information from a set
   ------------------------------------------------------


   function Search_Resource (
         My_Resources : in     Resources_Set;   
         Name         : in     Unbounded_String ) 
     return Resource_Ptr; 



   --------------------------------------------------------
   -- Functions to perform sort on a set
   --------------------------------------------------------



   procedure Same_Protocol_Control (
         My_Resources : in     Resources_Set;   
         A_Task       : in     Unbounded_String ); 





private

   type Resources_Set is new Generic_Resource_Set.Set with null record; 

end Resource_Set;




