------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool. 
-- This program provides services to automatically check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff, Jerome Legrand
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------






with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Unbounded_Strings;
use Unbounded_Strings;


with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Config;
use Config;

with Scheduler;



package Scheduler.Fixed_Priority is


   type Fixed_Priority_Scheduler is abstract new Generic_Scheduler with private; 


   type Fixed_Priority_Tcb is new Tcb with 
      record 
         Current_Priority : Priority_Range;  
         Is_Ready         : Boolean;  
      end record; 

   type Fixed_Priority_Tcb_Ptr is access all Fixed_Priority_Tcb'Class; 

   procedure Initialize (
         A_Tcb : in out Fixed_Priority_Tcb ); 

   ----------------------------------------

   type Fixed_Priority_Resource is new Shared_Resource with 
      record 
         Priority_Ceiling : Priority_Range;  
      end record; 

   type Fixed_Priority_Resource_Ptr is access all Fixed_Priority_Resource'Class; 

   ----------------------------------------

   function Build_Resource (
         My_Scheduler : in     Fixed_Priority_Scheduler; 
         A_Resource   :        Resource_Ptr              ) 
     return Shared_Resource_Ptr; 


   function Build_Tcb (
         My_Scheduler : in     Fixed_Priority_Scheduler; 
         A_Task       :        Task_Ptr                  ) 
     return Tcb_Ptr; 


   procedure Do_Election (
         My_Scheduler      : in out Fixed_Priority_Scheduler; 
         Si                : in out Scheduling_Information;   
         Result            : in out Scheduling_Sequence_Ptr;  
         Msg               : in out Unbounded_String;         
         Current_Time      : in     Natural;                  
         Processor_Name    : in     Unbounded_String;         
         My_Dependencies   : in     Tasks_Dependencies_Ptr;   
         With_Offsets      :        Boolean;                  
         With_Precedencies :        Boolean;                  
         With_Resources    :        Boolean;                  
        Event_To_Generate          : in Time_Unit_Event_Type_Boolean_Table;
         Elected           : in out Tasks_Range;              
         No_Task           : in out Boolean                   ); 


   procedure Compute_Response_Time (
         My_Scheduler   : in     Fixed_Priority_Scheduler; 
         My_Tasks       : in out Tasks_Set;                
         Processor_Name : in     Unbounded_String;         
         Msg            : in out Unbounded_String;         
         Response_Time  :    out Response_Table            ); 


   procedure Bound_On_Processor_Utilization (
         My_Scheduler   : in     Fixed_Priority_Scheduler; 
         My_Tasks       : in     Tasks_Set;                
         Processor_Name : in     Unbounded_String;         
         Result         : in out Double;                   
         Msg            : in out Unbounded_String          ); 


   procedure Set_Priority_According_To_Dm (
         My_Scheduler   : in     Fixed_Priority_Scheduler;                
         My_Tasks       : in out Tasks_Set;                               
         Processor_Name : in     Unbounded_String         := Empty_String ); 


   procedure Set_Priority_According_To_Rm (
         My_Scheduler   : in     Fixed_Priority_Scheduler;                
         My_Tasks       : in out Tasks_Set;                               
         Processor_Name : in     Unbounded_String         := Empty_String ); 


   procedure Check_Resource (
         My_Scheduler : in out Fixed_Priority_Scheduler; 
         Si           : in out Scheduling_Information;   
         Result       : in out Scheduling_Sequence_Ptr;  
         Current_Time : in     Natural;                  
         A_Tcb        : in     Tcb_Ptr;                  
         Is_Ready     :    out Boolean ;
  Event_To_Generate          : in Time_Unit_Event_Type_Boolean_Table                  ); 


   procedure Allocate_Resource (
         My_Scheduler : in out Fixed_Priority_Scheduler; 
         Si           : in out Scheduling_Information;   
         Result       : in out Scheduling_Sequence_Ptr;  
         Current_Time : in     Natural;                  
         A_Tcb        :        Tcb_Ptr          ;
  Event_To_Generate          : in Time_Unit_Event_Type_Boolean_Table         ); 


   procedure Release_Resource (
         My_Scheduler : in out Fixed_Priority_Scheduler; 
         Si           : in out Scheduling_Information;   
         Result       : in out Scheduling_Sequence_Ptr;  
         Current_Time : in     Natural;                  
         A_Tcb        :        Tcb_Ptr      ;
  Event_To_Generate          : in Time_Unit_Event_Type_Boolean_Table             ); 




   procedure Specific_Scheduler_Initialization (
         My_Scheduler   : in out Fixed_Priority_Scheduler; 
         Si             : in out Scheduling_Information;   
         Processor_Name : in     Unbounded_String;         
         My_Tasks       : in out Tasks_Set;                
         My_Resources   : in out Resources_Set;            
         My_Buffers     : in out Buffers_Set;              
         My_Messages    : in     Messages_Set;             
         Msg            : in out Unbounded_String          ); 


   procedure Compute_Ceiling_Of_Resources (
         My_Scheduler   : in out Fixed_Priority_Scheduler; 
         Si             : in out Scheduling_Information;   
         Processor_Name : in     Unbounded_String;         
         My_Tasks       : in out Tasks_Set;                
         My_Resources   : in out Resources_Set             ); 


   procedure Dispatched_Change_Current_Priority (
         My_Scheduler : in out Fixed_Priority_Scheduler; 
         A_Tcb        : in     Fixed_Priority_Tcb_Ptr;   
         New_Priority :        Priority_Range            ); 


   procedure Change_Current_Priority (
         My_Scheduler : in out Fixed_Priority_Scheduler'Class; 
         A_Tcb        : in     Fixed_Priority_Tcb_Ptr;         
         New_Priority :        Priority_Range                  ); 


   function Compute_Bound_On_Task_Blocking_Time (
         My_Scheduler : in     Fixed_Priority_Scheduler; 
         My_Tasks     : in     Tasks_Set;                
         My_Resources : in     Resources_Set;            
         A_Task       : in     Unbounded_String          ) 
     return Natural; 




private

   type Fixed_Priority_Scheduler is abstract new Generic_Scheduler with 
      record 
         Used_Resource : Boolean := False;  
      end record; 


end Scheduler.Fixed_Priority ;


