------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff, Jerome Legrand
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Io_Tools;
use Io_Tools;
with Ada.Io_Exceptions;
use Ada.Io_Exceptions;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;
with Ada.Strings.Unbounded.Text_Io;
use Ada.Strings.Unbounded.Text_Io;
with Unbounded_Strings;
use Unbounded_Strings;

with Ada.Exceptions;
use Ada.Exceptions;

with Translate;
use Translate;

with  Scheduler_Io;
use  Scheduler_Io;
with Text_Io;
use Text_Io;

with Statements;
use Statements;
use Statements.Sets_Type_Package;
with Simulations;
use Simulations;

with Parameters;
use Parameters;
use Parameters.User_Defined_Parameters_Table_Package;

with Parser;
use Parser;


with Task_Set;
use Task_Set;




package body Scheduler.Parametrics is




   procedure Initialize (
         A_Scheduler : in out Parametric_Scheduler ) is 
   begin
      Reset(A_Scheduler);
      A_Scheduler.Name:= Parametric_Protocol;
      A_Scheduler.String_Behavior:= new Unbounded_String;
   end Initialize;


   function Copy (
         A_Scheduler : in     Parametric_Scheduler ) 
     return Scheduler_Ptr is 

      Ptr : Parametric_Scheduler_Ptr;  

   begin

      Ptr:= new Parametric_Scheduler;

      Ptr.Name:=A_Scheduler.Name;
      Ptr.Is_Preemptive:=A_Scheduler.Is_Preemptive;
      Ptr.Quantum:=A_Scheduler.Quantum;
      Ptr.Previously_Elected:=A_Scheduler.Previously_Elected;

      Ptr.String_Behavior:= new Unbounded_String;
      Ptr.String_Behavior.All:=A_Scheduler.String_Behavior.All;
      Ptr.Root_Statement_Pointer:=A_Scheduler.Root_Statement_Pointer;
      Ptr.Variables_Table:=A_Scheduler.Variables_Table;

      return Scheduler_Ptr(Ptr);

   end Copy;


   ---------------------------------------------------------------------
   --  Standard procedures and functions of scheduler class
   ----------------------------------------------------------------------



   procedure Bound_On_Processor_Utilization (
         My_Scheduler   : in     Parametric_Scheduler; 
         My_Tasks       : in     Tasks_Set;            
         Processor_Name : in     Unbounded_String;     
         Result         : in out Double;               
         Msg            : in out Unbounded_String      ) is 

   begin
      raise Invalid_Scheduler;
   end Bound_On_Processor_Utilization;


   procedure Compute_Response_Time (
         My_Scheduler   : in     Parametric_Scheduler; 
         My_Tasks       : in out Tasks_Set;            
         Processor_Name : in     Unbounded_String;     
         Msg            : in out Unbounded_String;     
         Response_Time  :    out Response_Table        ) is 

   begin
      raise Invalid_Scheduler;
   end Compute_Response_Time;


   function Get_String_Behavior (
         My_Scheduler : in     Parametric_Scheduler ) 
     return Unbounded_String is 

   begin
      return My_Scheduler.String_Behavior.All;
   end Get_String_Behavior;


   function Get_String_Behavior (
         My_Scheduler : in     Parametric_Scheduler ) 
     return Unbounded_String_Ptr is 

   begin
      return My_Scheduler.String_Behavior;
   end Get_String_Behavior;


   procedure Put (
         My_Scheduler : in     Parametric_Scheduler ) is 
   begin
      Put(Generic_Scheduler(My_Scheduler));
      Put_Line("Behavior file  name : " & My_Scheduler.Behavior_File_Name);
      Put_Line("Behavior file  : " & My_Scheduler.String_Behavior.All);
      Put(My_Scheduler.Root_Statement_Pointer);
      Put(My_Scheduler.Variables_Table);
   end Put;


   procedure Set_Behavior_File_Name (
         My_Scheduler : in out Parametric_Scheduler; 
         To_Set       : in     Unbounded_String      ) is 
   begin
      My_Scheduler.Behavior_File_Name:=To_Set;
   end Set_Behavior_File_Name;


   function Get_Behavior_File_Name (
         My_Scheduler : in     Parametric_Scheduler ) 
     return Unbounded_String is 
   begin
      return My_Scheduler.Behavior_File_Name;
   end Get_Behavior_File_Name;


   procedure Set_String_Behavior (
         My_Scheduler : in out Parametric_Scheduler; 
         To_Set       : in     Unbounded_String      ) is 
   begin
      My_Scheduler.String_Behavior:= new Unbounded_String;
      My_Scheduler.String_Behavior.All:=To_Set;
   end Set_String_Behavior;


   procedure Check_Before_Scheduling (
         My_Scheduler   : in     Parametric_Scheduler; 
         My_Tasks       : in     Tasks_Set;            
         Processor_Name : in     Unbounded_String      ) is 

   begin
      null;
   end Check_Before_Scheduling;



   procedure Do_Election (
         My_Scheduler      : in out Parametric_Scheduler;               
         Si                : in out Scheduling_Information;             
         Result            : in out Scheduling_Sequence_Ptr;            
         Msg               : in out Unbounded_String;                   
         Current_Time      : in     Natural;                            
         Processor_Name    : in     Unbounded_String;                   
         My_Dependencies   : in     Tasks_Dependencies_Ptr;             
         With_Offsets      : in     Boolean;                            
         With_Precedencies : in     Boolean;                            
         With_Resources    : in     Boolean;                            
         Event_To_Generate : in     Time_Unit_Event_Type_Boolean_Table; 
         Elected           : in out Tasks_Range;                        
         No_Task           : in out Boolean                             ) is 


      Return_Val             : Expression_Ptr;  
      Current                : Statement_Ptr;  
      Election_Value         : Simulation_Value_Ptr;  
      Current_Selected_Value : Simulation_Value_Ptr;  
      I                      : Tasks_Range          := 0;  
      Var                    : Variables_Range;  


   begin


      -- By default, no task are elected : it's a way
      -- to safer schedule a parametric scheduler without 
      -- a correctly designed code (election part of it)
      --
      No_Task:=True;



      for I in 0..Si.Number_Of_Tasks -1 loop

         Var:=Find_Variable(My_Scheduler.Variables_Table,
            To_Unbounded_String("tasks.activation_number"));
         My_Scheduler.Variables_Table.Entries(Var).Simulation.
            Integer_Table_Value(Simulations_Range(I)):=
            Integer(Si.Tcbs(I).Activation);

      end loop;

      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("tasks.rest_of_capacity"));
      My_Scheduler.Variables_Table.Entries(Var).Simulation.
         Integer_Table_Value(Simulations_Range(My_Scheduler.
            Previously_Elected)):=
         Integer(Si.Tcbs(My_Scheduler.Previously_Elected).
         Rest_Of_Capacity);


      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("tasks.used_cpu"));
      My_Scheduler.Variables_Table.Entries(Var).Simulation.
         Integer_Table_Value(Simulations_Range(My_Scheduler.
            Previously_Elected)):=
         Integer(Si.Tcbs(My_Scheduler.Previously_Elected).Used_Cpu);


      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("previously_elected"));
      My_Scheduler.Variables_Table.Entries(Var).Simulation.
         Integer_Value:=Integer(My_Scheduler.Previously_Elected);


      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("simulation_time"));
      My_Scheduler.Variables_Table.Entries(Var).Simulation.
         Integer_Value:=Integer(Current_Time);


      -- check for each task is they are ready or not
      --
      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("tasks.ready"));


      for I in 0..Si.Number_Of_Tasks -1 loop
         My_Scheduler.Variables_Table.Entries(Var).Simulation.
            Boolean_Table_Value(Simulations_Range(I)):=
            False;
      end loop;


      for I in 0..Si.Number_Of_Tasks -1 loop
         if (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) then
            if (Si.Tcbs(I).Wake_Up_Time <= Current_Time) and
                  (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                  then
               if (With_Offsets=False) or
                     Check_Offset(Si.Tcbs(I), Current_Time)
                     then
                  if (With_Precedencies=False) or
                        Check_Precedencies(Si, My_Dependencies,
                        Current_Time,
                        Si.Tcbs(I))
                        then

                     My_Scheduler.Variables_Table.Entries(Var).
                        Simulation.
                        Boolean_Table_Value(Simulations_Range(I)):=
                        True;

                  end if;
               end if;
            end if;
         end if;

      end loop;



      -- Now, do "priority" section
      --
      Current := My_Scheduler.Root_Statement_Pointer(Priority);
      Dispatch(Current, Priority, My_Scheduler.Variables_Table, Msg);


      -- Do "election" section
      --

      -- A election is done in 3 steps :
      --        1) we check and retrieve return information
      --        2) tasks queues should be updated
      --        3) we find the running task
      --
      Current := My_Scheduler.Root_Statement_Pointer(Election);

      -- Only one type of statement is allowed on "election" section :
      -- a return statement
      --
      if Current/=null then
         if Current.All not in Return_Statement'Class
               then
            Raise_Exception (Statement_Error'Identity,
               "Election section, line " & Current.Line_Number'Img &
               ", return statement expected");
         end if;


         -- Do return statement
         --

         -- The return statement should be :
         --      - either a min/max_to_index expression composed of
         --        a array variable only. Boolean/double/string array variable are forbidden
         --      - or an integer constant
         --      - or an integer variable
         -- 
         Return_Val:=Return_Statement_Ptr(Current).Return_Value;




         ------------------------------------------------------------
         -- The return statement is a constant or a variable
         ------------------------------------------------------------


         if (Return_Val.Expression_Type=Variable_Type) or (Return_Val.
               Expression_Type=Constant_Type) then


            if (Return_Val.Expression_Type=Constant_Type)
                  then
               if Constant_Expr_Ptr(Return_Val).Value.Ptype/=
                     Simulation_Integer
                     then
                  Raise_Exception (Statement_Error'Identity,
                     "Election section, line " & Current.Line_Number'Img &
                     ", invalid constant expression type referenced in a return statement");
               else
                  if (Constant_Expr_Ptr(Return_Val).Value.Integer_Value<
                        Integer(Tasks_Range'First)) or
                        (Constant_Expr_Ptr(Return_Val).Value.
                        Integer_Value>
                        Integer(Tasks_Range'
                           Last)) then
                     Raise_Exception (Statement_Error'Identity,
                        "Election section, line " & Current.Line_Number'
                        Img &
                        ", invalid range constant type referenced in a return statement");
                  end if;

                  I:=Tasks_Range(Constant_Expr_Ptr(Return_Val).Value.
                     Integer_Value);
               end if;

            else

               if   (Return_Val.Expression_Type=Variable_Type)
                     then


                  Var:=Find_Variable(My_Scheduler.Variables_Table,
                     Variable_Expr_Ptr(Return_Val).Name);


                  if My_Scheduler.Variables_Table.Entries(Var).
                        Simulation.
                        Ptype/=Simulation_Integer
                        then
                     Raise_Exception (Statement_Error'Identity,
                        "Election section, line " & Current.Line_Number'
                        Img &
                        ", invalid variable expression type referenced in a return statement");
                  end if;

                  if (My_Scheduler.Variables_Table.Entries(Var).
                        Simulation.Integer_Value<Integer(
                           Simulations_Range'First)) or
                        (My_Scheduler.Variables_Table.Entries(Var).
                        Simulation.Integer_Value>Integer(
                           Tasks_Range'
                           Last)) then
                     Raise_Exception (Statement_Error'Identity,
                        "Election section, line " & Current.Line_Number'
                        Img &
                        ", invalid range variable type referenced in a return statement");
                  end if;
                  I:=Tasks_Range(My_Scheduler.Variables_Table.Entries(Var).
                     Simulation.Integer_Value);
               end if;
            end if;


            if (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) then
               if (Si.Tcbs(I).Wake_Up_Time <= Current_Time) and
                     (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                     then
                  if (With_Offsets=False) or
                        Check_Offset(Si.Tcbs(I), Current_Time)
                        then
                     if (With_Precedencies=False) or
                           Check_Precedencies(Si, My_Dependencies,
                           Current_Time,
                           Si.Tcbs(I))
                           then
                        Elected:=I;
                        No_Task:=False;

                     end if;
                  end if;
               end if;
            end if;





            ------------------------------------------------------------
            -- Now, let see the case when the return expression is a max/min_to_index
            ------------------------------------------------------------
         else
            if (Return_Val.Expression_Type=Unary_Type) then
               if (Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index) or (
                     Unary_Expr_Ptr(Return_Val).Ope=Max_To_Index) then



                  if (Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index) then
                     if (Unary_Expr_Ptr(Return_Val).Value.
                           Expression_Type/=
                           Array_Variable_Type)
                           then
                        Raise_Exception (Statement_Error'Identity,
                           "Election section, line " & Current.
                           Line_Number'
                           Img &
                           ", invalid min_to_index expression in return statement");
                     end if;

                     if (Array_Variable_Expr_Ptr(Unary_Expr_Ptr(
                                 Return_Val).
                              Value).
                           Variable_Type
                           /=Simulation_Array_Integer)
                           and (Array_Variable_Expr_Ptr(
                              Unary_Expr_Ptr(
                                 Return_Val).
                              Value).
                           Variable_Type
                           /=Simulation_Array_Double)

                           then
                        Raise_Exception (Statement_Error'Identity,
                           "Election section, line " & Current.
                           Line_Number'
                           Img &
                           ", min_to_index expression in return statement should be an integer array variable");
                     end if;
                  end if;
               end if;

               if (Return_Val.Expression_Type=Unary_Type) then
                  if (Unary_Expr_Ptr(Return_Val).Ope=Max_To_Index) then
                     if (Unary_Expr_Ptr(Return_Val).Value.
                           Expression_Type/=
                           Array_Variable_Type)
                           then
                        Raise_Exception (Statement_Error'Identity,
                           "Election section, line " & Current.
                           Line_Number'
                           Img &
                           ", invalid max_to_index expression in return statement");
                     end if;


                     if (Array_Variable_Expr_Ptr(Unary_Expr_Ptr(
                                 Return_Val).
                              Value).
                           Variable_Type
                           /=Simulation_Array_Integer)
                           then
                        Raise_Exception (Statement_Error'Identity,
                           "Election section, line " & Current.
                           Line_Number'
                           Img &
                           ", max_to_index expression in return statement should be an integer array variable");
                     end if;
                  end if;
               end if;

               -----------------------------------------------------
               -- All the code is now done : we do the last
               -- operation : election !!!
               -----------------------------------------------------

               Election_Value:=Value_Of(Array_Variable_Expr(
                     Unary_Expr_Ptr(
                        Return_Val).Value.All), My_Scheduler.
                  Variables_Table);


               if (Election_Value.Ptype=Simulation_Array_Double) then
                  Current_Selected_Value:= new Simulation_Value (
                     Simulation_Double);
                  if Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index then
                     Current_Selected_Value.Double_Value:=Double'Last;
                  else
                     Current_Selected_Value.Double_Value:=Double'First;
                  end if;
               else
                  Current_Selected_Value:= new Simulation_Value (
                     Simulation_Integer);
                  if Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index then
                     Current_Selected_Value.Integer_Value:=Integer'Last;
                  else
                     Current_Selected_Value.Integer_Value:=Integer'First;
                  end if;
               end if;


               -- Now, loop on the array variable to find the appropriate task to be
               -- elected
               --

               loop
                  if (Si.Tcbs(I).Tsk.Cpu_Name = Processor_Name) then
                     if (Si.Tcbs(I).Wake_Up_Time <= Current_Time) and
                           (Si.Tcbs(I).Rest_Of_Capacity /= 0)
                           then
                        if (With_Offsets=False) or
                              Check_Offset(Si.Tcbs(I), Current_Time)
                              then
                           if (With_Precedencies=False) or
                                 Check_Precedencies(Si, My_Dependencies,
                                 Current_Time,
                                 Si.Tcbs(I))
                                 then

                              if (Election_Value.Ptype=
                                    Simulation_Array_Integer) then

                                 if ((Unary_Expr_Ptr(Return_Val).Ope=
                                          Min_To_Index) and
                                       (Election_Value.
                                          Integer_Table_Value(
                                             Simulations_Range(I)) <
                                          Current_Selected_Value.
                                          Integer_Value))
                                       or
                                       ((Unary_Expr_Ptr(Return_Val).Ope=
                                          Max_To_Index) and
                                       (Election_Value.
                                          Integer_Table_Value(
                                             Simulations_Range(I)) >
                                          Current_Selected_Value.
                                          Integer_Value))

                                       then
                                    Elected:=I;
                                    Current_Selected_Value.Integer_Value:=
                                       Election_Value.
                                       Integer_Table_Value(
                                       Simulations_Range(I));
                                 end if;
                              else
                                 if ((Unary_Expr_Ptr(Return_Val).Ope=
                                          Min_To_Index) and
                                       (Election_Value.Double_Table_Value(
                                             Simulations_Range(I)) <
                                          Current_Selected_Value.
                                          Double_Value))
                                       or
                                       ((Unary_Expr_Ptr(Return_Val).Ope=
                                          Max_To_Index) and
                                       (Election_Value.Double_Table_Value(
                                             Simulations_Range(I)) >
                                          Current_Selected_Value.
                                          Double_Value))

                                       then
                                    Elected:=I;
                                    Current_Selected_Value.Double_Value:=
                                       Election_Value.
                                       Double_Table_Value(
                                       Simulations_Range(I));
                                 end if;


                              end if;
                           end if;
                        end if;
                     end if;
                  end if;


                  I:=I+1;
                  exit when Si.Tcbs(I)=null;
               end loop;



               if (Election_Value.Ptype=
                     Simulation_Array_Integer) then

                  if ( (Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index) and (
                           Current_Selected_Value.Integer_Value/=Integer'
                           Last)  )
                        or ( (Unary_Expr_Ptr(Return_Val).Ope=
                           Max_To_Index) and (
                           Current_Selected_Value.Integer_Value/=Integer'
                           First)  )
                        then
                     No_Task:=False;
                  end if;
               else
                  if ( (Unary_Expr_Ptr(Return_Val).Ope=Min_To_Index) and (
                           Current_Selected_Value.Double_Value/=Double'
                           Last)  )
                        or ( (Unary_Expr_Ptr(Return_Val).Ope=
                           Max_To_Index) and (
                           Current_Selected_Value.Double_Value/=Double'
                           First)  )
                        then
                     No_Task:=False;
                  end if;
               end if;


               -- End of return statement
               --

               ------------------------------------------------------------
               -- Others expression in a return statement is a mistake 
               ------------------------------------------------------------
            else

               Raise_Exception (Statement_Error'Identity,
                  "Election section, line " & Current.Line_Number'Img &
                  ", invalid expression referenced in a return statement");


            end if;
         end if;


      end if;


      -- After election section, we
      -- save some task run time information
      --
      Save_Read_Write_Parametric_Variables(My_Scheduler, Si);





   end Do_Election;



   -- Do "start" section statements
   --
   procedure Specific_Scheduler_Initialization (
         My_Scheduler   : in out Parametric_Scheduler;   
         Si             : in out Scheduling_Information; 
         Processor_Name : in     Unbounded_String;       
         My_Tasks       : in out Tasks_Set;              
         My_Resources   : in out Resources_Set;          
         My_Buffers     : in out Buffers_Set;            
         My_Messages    : in     Messages_Set;           
         Msg            : in out Unbounded_String        ) is 


      Current : Statement_Ptr;  
      Var     : Variables_Range;  


   begin


      -- Read all parametric files before simulation
      --
      begin
         My_Scheduler.String_Behavior:= new Unbounded_String;
         My_Scheduler.String_Behavior.All:=Read_Sequential_File(
            My_Scheduler.Behavior_File_Name);


      exception
         when Ada.Io_Exceptions.Name_Error =>
            Raise_Exception(Parametric_File_Error'Identity, To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(Current_Language))  &
               ", Name_Error");
         when Ada.Io_Exceptions.Status_Error =>
            Raise_Exception(Parametric_File_Error'Identity, To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Status_Error");
         when Ada.Io_Exceptions.Mode_Error   =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Mode_Error");
         when Ada.Io_Exceptions.Use_Error     =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(Current_Language)) &
               ", Use_Error");
         when Ada.Io_Exceptions.Device_Error  =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(Current_Language)) &
               ", Device_Error");
         when Ada.Io_Exceptions.End_Error    =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", End_Error");
         when Ada.Io_Exceptions.Data_Error   =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Data_Error");
         when Ada.Io_Exceptions.Layout_Error  =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) &
               ", Layout_Error");
         when others =>
            Raise_Exception(Parametric_File_Error'Identity,To_String(
                  Lb_File(Current_Language) & Lb_Colon &
                  My_Scheduler.Behavior_File_Name & Lb_Comma &
                  Lb_Can_Not_Open_File(
                     Current_Language)) );
      end;



      -- We should check the scheduler behavior syntax ...
      --
      Open_Input(My_Scheduler.String_Behavior.All);
      Parser.Init_Parser(My_Tasks);
      Parser.Yyparse;



      -- Store syntax tree and compiling information
      --
      My_Scheduler.Root_Statement_Pointer:=
         Parser.Root_Statement_Pointer;

      My_Scheduler.Variables_Table:=
         Parser.Variables_Table;

      My_Scheduler.Sets_Table:=
         Parser.Sets_Table;

      Current := My_Scheduler.Root_Statement_Pointer (Start);



      -- Finish simulation data initialization
      --

      -- Load generic variables
      --
      Load_Parametric_Variables (
         My_Scheduler.Variables_Table,
         My_Messages);
      Load_Parametric_Variables (
         My_Scheduler.Variables_Table,
         My_Buffers);
      Load_Parametric_Variables (
         My_Scheduler.Variables_Table,
         My_Resources);

      -- Load parametric variable before executing sections
      --
      Load_Read_Write_Parametric_Variables(My_Scheduler, Si);


      -- Set variable : "nb_tasks"
      --
      Var:=Find_Variable(My_Scheduler.Variables_Table,
         To_Unbounded_String("nb_tasks"));
      My_Scheduler.Variables_Table.Entries(Var).Simulation.
         Integer_Value:=Integer(Get_Number_Of_Task_From_Processor(
            My_Tasks, Processor_Name));





      -- Now : run "start" statements
      --
      Dispatch(Current, Start, My_Scheduler.Variables_Table, Msg);


      -- After start section, we
      -- save some task run time information
      --
      Save_Read_Write_Parametric_Variables(My_Scheduler, Si);


   end Specific_Scheduler_Initialization;


   procedure Compute_Activation_Time (
         My_Scheduler : in     Parametric_Scheduler;   
         Si           : in out Scheduling_Information; 
         Elected      : in     Tasks_Range;            
         Value        : in out Natural                 ) is 


      Set_Id           : Sets_Range;  
      Parametric_Delay : Simulation_Value_Ptr;  

   begin

      Set_Id:=Find_Set(My_Scheduler.Sets_Table,
         Parametric_Task_Ptr(Si.Tcbs(Elected).Tsk).Activation_Rule);

      Parametric_Delay := Value_Of (My_Scheduler.
         Sets_Table.Entries(
            Set_Id).Set_Value.All,
         My_Scheduler.Variables_Table);
      --     Parametric_Delay := Value_Of (Array_Variable_Expr(My_Scheduler.
      --            Sets_Table.Entries(
      --               Set_Id).Set_Value.all),
      --         My_Scheduler.Variables_Table);


      if (Parametric_Delay.Ptype /= Simulation_Integer)
            and  (Parametric_Delay.Ptype /=
            Simulation_Array_Integer) then
         Raise_Exception(Type_Error'Identity,
            "Task activation rule, activation rule " &
            To_String(Parametric_Task_Ptr(Si.Tcbs(Elected).Tsk).
               Activation_Rule)  &
            To_String(Lb_Comma & Lb_Uncompatible_Type_Error(
                  Current_Language) ));
      end if;



      -- If it's a vector, return the corresponding index
      -- else, just return the computed integer
      --
      if Parametric_Delay.Ptype=Simulation_Integer
            then
         Value:=Parametric_Delay.Integer_Value;
      else
         Value:=Parametric_Delay.Integer_Table_Value(Simulations_Range(
               Elected));
      end if;


   end Compute_Activation_Time;



   procedure Utilization_Factor_Feasibility_Test (
         My_Scheduler   : in     Parametric_Scheduler; 
         My_Tasks       : in     Tasks_Set;            
         Processor_Name : in     Unbounded_String;     
         Result         : in out Unbounded_String      ) is 

   begin
      Result := Result &  Lb_Compute_Scheduling_Error_19(Current_Language) &
         Unbounded_Lf;
   end Utilization_Factor_Feasibility_Test;





   procedure Load_Read_Write_Parametric_Variables (
         My_Scheduler : in out Parametric_Scheduler;  
         Si           : in     Scheduling_Information ) is 

      Var : Variables_Range;  

   begin


      for J in 0..Si.Number_Of_Tasks -1 loop


         -- "priority"
         --
         Var:=Find_Variable(Variables_Table,
            To_Unbounded_String("tasks.priority"));
         My_Scheduler.Variables_Table.Entries(Var).Simulation.
            Integer_Table_Value(Simulations_Range(J)):=
            Integer(Si.Tcbs(J).Tsk.Priority);


         -- "deadline"
         --
         Var:=Find_Variable(Variables_Table,
            To_Unbounded_String("tasks.deadline"));
         My_Scheduler.Variables_Table.Entries(Var).Simulation.
            Integer_Table_Value(Simulations_Range(J)):=
            Integer(Si.Tcbs(J).Tsk.Deadline);


         -- "start_time"
         --
         Var:=Find_Variable(Variables_Table,
            To_Unbounded_String("tasks.start_time"));
         My_Scheduler.Variables_Table.Entries(Var).Simulation.
            Integer_Table_Value(Simulations_Range(J)):=
            Integer(Si.Tcbs(J).Tsk.Start_Time);


         -- "processor_name"
         --
         Var:=Find_Variable(Variables_Table,
            To_Unbounded_String("tasks.processor_name"));
         My_Scheduler.Variables_Table.Entries(Var).Simulation.
            String_Table_Value(Simulations_Range(J)):=
            Si.Tcbs(J).Tsk.Cpu_Name;


         -- "name"
         --
         Var:=Find_Variable(Variables_Table,
            To_Unbounded_String("tasks.name"));
         My_Scheduler.Variables_Table.Entries(Var).Simulation.
            String_Table_Value(Simulations_Range(J)):=
            Si.Tcbs(J).Tsk.Name;


         -- "capacity"
         --
         Var:=Find_Variable(Variables_Table,
            To_Unbounded_String("tasks.capacity"));
         My_Scheduler.Variables_Table.Entries(Var).Simulation.
            Integer_Table_Value(Simulations_Range(J)):=
            Integer(Si.Tcbs(J).Tsk.Capacity);


         -- "blocking_time"
         --
         Var:=Find_Variable(Variables_Table,
            To_Unbounded_String("tasks.blocking_time"));
         My_Scheduler.Variables_Table.Entries(Var).Simulation.
            Integer_Table_Value(Simulations_Range(J)):=
            Integer(Si.Tcbs(J).Tsk.Blocking_Time);


         -- "period"
         --
         Var:=Find_Variable(Variables_Table,
            To_Unbounded_String("tasks.period"));
         case Si.Tcbs(J).Tsk.Task_Type is
            when Aperiodic_Type =>
               My_Scheduler.Variables_Table.Entries(Var).Simulation.
                  Integer_Table_Value(Simulations_Range(J)):=0;
            when others =>
               My_Scheduler.Variables_Table.Entries(Var).Simulation.
                  Integer_Table_Value(Simulations_Range(J)):=
                  Integer(Periodic_Task_Ptr(Si.Tcbs(J).Tsk).Period);
         end case;


         -- "jitter"
         --
         Var:=Find_Variable(Variables_Table,
            To_Unbounded_String("tasks.jitter"));
         case Si.Tcbs(J).Tsk.Task_Type is
            when Aperiodic_Type =>
               My_Scheduler.Variables_Table.Entries(Var).Simulation.
                  Integer_Table_Value(Simulations_Range(J)):=0;
            when others =>
               My_Scheduler.Variables_Table.Entries(Var).Simulation.
                  Integer_Table_Value(Simulations_Range(J)):=
                  Integer(Periodic_Task_Ptr(Si.Tcbs(J).Tsk).Jitter);
         end case;



         -- user defined task's parameter :
         --

         for I in 0..Si.Tcbs(J).Tsk.Param.Nb_Entries-1 loop

            -- Find the user defined variable
            -- and set its value
            --
            for K in 0..Variables_Table.Nb_Entries-1 loop
               if My_Scheduler.Variables_Table.Entries(K).Variable.Name=
                     To_Unbounded_String("tasks.") & Si.Tcbs(J).
                     Tsk.Param.
                     Entries(
                     I).Name
                     then

                  -- Now, set the value
                  --
                  if Si.Tcbs(J).Tsk.Param.Entries(I).Ptype=
                        Boolean_Parameter then
                     My_Scheduler.Variables_Table.Entries(K).
                        Simulation.
                        Boolean_Table_Value(Simulations_Range(J)):=
                        Boolean(Si.Tcbs(J).Tsk.
                        Param.Entries(I).Boolean_Value);
                  end if;
                  if Si.Tcbs(J).Tsk.Param.Entries(I).Ptype=
                        Integer_Parameter then
                     My_Scheduler.Variables_Table.Entries(K).
                        Simulation.
                        Integer_Table_Value(Simulations_Range(J)):=
                        Integer(Si.Tcbs(J).Tsk.
                        Param.Entries(I).Integer_Value);
                  end if;
                  if Si.Tcbs(J).Tsk.Param.Entries(I).Ptype=
                        Double_Parameter then
                     My_Scheduler.Variables_Table.Entries(K).
                        Simulation.
                        Double_Table_Value(Simulations_Range(J)):=Double(
                        Si.Tcbs(J).Tsk.
                        Param.Entries(I).Double_Value);
                  end if;
                  if Si.Tcbs(J).Tsk.Param.Entries(I).Ptype=
                        String_Parameter then
                     My_Scheduler.Variables_Table.Entries(K).
                        Simulation.
                        String_Table_Value(Simulations_Range(J)):=
                        Si.Tcbs(J).Tsk.
                        Param.Entries(I).String_Value;
                  end if;
               end if;
            end loop;
         end loop;

      end loop;


   end Load_Read_Write_Parametric_Variables;





   procedure Save_Read_Write_Parametric_Variables (
         My_Scheduler : in     Parametric_Scheduler;  
         Si           : in out Scheduling_Information ) is 

      Var : Variables_Range;  

   begin

      -- "priority"
      --
      Var:=Find_Variable(Variables_Table,
         To_Unbounded_String("tasks.priority"));
      for J in 0..Si.Number_Of_Tasks -1 loop
         begin
            Si.Tcbs(J).Tsk.Priority:=Priority_Range(My_Scheduler.
               Variables_Table.Entries(
                  Var).Simulation.
               Integer_Table_Value(Simulations_Range(J)));
         exception
            when Constraint_Error=>
               Si.Tcbs(J).Tsk.Priority:=Priority_Range'First;
         end;
      end loop;


      -- "deadline"
      --
      Var:=Find_Variable(Variables_Table,
         To_Unbounded_String("tasks.deadline"));
      for J in 0..Si.Number_Of_Tasks -1 loop
         Si.Tcbs(J).Tsk.Deadline:=My_Scheduler.Variables_Table.Entries(
            Var).Simulation.
            Integer_Table_Value(Simulations_Range(J));
      end loop;


      -- "start_time"
      --
      Var:=Find_Variable(Variables_Table,
         To_Unbounded_String("tasks.start_time"));
      for J in 0..Si.Number_Of_Tasks -1 loop
         Si.Tcbs(J).Tsk.Start_Time:=My_Scheduler.Variables_Table.Entries(
            Var).Simulation.
            Integer_Table_Value(Simulations_Range(J));
      end loop;


      -- "capacity"
      --
      Var:=Find_Variable(Variables_Table,
         To_Unbounded_String("tasks.capacity"));
      for J in 0..Si.Number_Of_Tasks -1 loop
         Si.Tcbs(J).Tsk.Capacity:=My_Scheduler.Variables_Table.Entries(
            Var).
            Simulation.
            Integer_Table_Value(Simulations_Range(J));
      end loop;


      -- "blocking_time"
      --
      Var:=Find_Variable(Variables_Table,
         To_Unbounded_String(
            "tasks.blocking_time"));
      for J in 0..Si.Number_Of_Tasks -1 loop
         Si.Tcbs(J).Tsk.Blocking_Time:=My_Scheduler.Variables_Table.
            Entries(Var).
            Simulation.
            Integer_Table_Value(Simulations_Range(J));
      end loop;


      -- "period"
      --
      Var:=Find_Variable(Variables_Table,
         To_Unbounded_String("tasks.period"));
      for J in 0..Si.Number_Of_Tasks -1 loop
         case Si.Tcbs(J).Tsk.Task_Type is
            when Aperiodic_Type =>
               null;
            when others =>
               Periodic_Task_Ptr(Si.Tcbs(J).Tsk).Period:=My_Scheduler.
                  Variables_Table.
                  Entries(Var).Simulation.
                  Integer_Table_Value(Simulations_Range(J));
         end case;
      end loop;


      -- "jitter"
      --
      Var:=Find_Variable(Variables_Table,
         To_Unbounded_String("tasks.jitter"));
      for J in 0..Si.Number_Of_Tasks -1 loop
         case Si.Tcbs(J).Tsk.Task_Type is
            when Aperiodic_Type =>
               null;
            when others =>
               Periodic_Task_Ptr(Si.Tcbs(J).Tsk).Jitter:=My_Scheduler.
                  Variables_Table.
                  Entries(Var).Simulation.
                  Integer_Table_Value(Simulations_Range(J));
         end case;
      end loop;


      -- user defined task's parameter :
      --
      for J in 0..Si.Number_Of_Tasks -1 loop
         for I in 0..Si.Tcbs(J).Tsk.Param.
               Nb_Entries-1 loop

            -- Find the user defined variable
            -- and set its value
            --
            for K in 0..Variables_Table.
                  Nb_Entries-1 loop
               if My_Scheduler.
                     Variables_Table.
                     Entries(K).Variable.
                     Name=
                     To_Unbounded_String(
                     "tasks.") & Si.Tcbs(J).
                     Tsk.Param.
                     Entries(
                     I).Name
                     then

                  -- Now, set the value
                  --
                  if Si.Tcbs(J).Tsk.Param.
                        Entries(I).Ptype=
                        Boolean_Parameter then
                     Si.Tcbs(J).
                        Tsk.
                        Param.Entries(I).
                        Boolean_Value:=My_Scheduler.
                        Variables_Table.
                        Entries(K).
                        Simulation.
                        Boolean_Table_Value(
                        Simulations_Range(J));
                  end if;
                  if Si.Tcbs(J).Tsk.Param.
                        Entries(I).Ptype=
                        Integer_Parameter then
                     Si.Tcbs(J).
                        Tsk.
                        Param.Entries(I).
                        Integer_Value:=My_Scheduler.
                        Variables_Table.
                        Entries(K).
                        Simulation.
                        Integer_Table_Value(
                        Simulations_Range(J));
                  end if;
                  if Si.Tcbs(J).Tsk.Param.
                        Entries(I).Ptype=
                        Double_Parameter then
                     Si.Tcbs(J).Tsk.
                        Param.Entries(I).
                        Double_Value:=My_Scheduler.
                        Variables_Table.
                        Entries(K).
                        Simulation.
                        Double_Table_Value(
                        Simulations_Range(J));
                  end if;
                  if Si.Tcbs(J).Tsk.Param.
                        Entries(I).Ptype=
                        String_Parameter then
                     Si.Tcbs(J).Tsk.
                        Param.Entries(I).
                        String_Value:=My_Scheduler.
                        Variables_Table.
                        Entries(K).
                        Simulation.
                        String_Table_Value(
                        Simulations_Range(J));
                  end if;
               end if;
            end loop;
         end loop;
      end loop;

   end Save_Read_Write_Parametric_Variables;




end Scheduler.Parametrics ;


