------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a free real time scheduling tool.
-- This program provides services to automatically check temporal constraints
-- of real time tasks.
--
-- Copyright (C) 2002-2007   Frank Singhoff, Jerome Legrand
-- Cheddar is developed by the LISYC Team, University of Brest
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------


with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

with Config;
use Config;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Double_Util;
use Double_Util;

with Tasks;
use Tasks;
with Buffers;
use Buffers;

with Task_Set;
use Task_Set;
with Resource_Set;
use Resource_Set;

with Event_Analyzer_Set;
use Event_Analyzer_Set;
use Event_Analyzer_Set.Generic_Event_Analyzer_Set;


with Tables;
with Indexed_Tables;
with Lists;
with Unchecked_Deallocation;

with Queueing_System;
use Queueing_System;


package Scheduling_Sequences is



   --------------------------------------------------------
   --------------------------------------------------------
   --
   -- Definition of scheduling sequence type and its services
   --
   --------------------------------------------------------
   --------------------------------------------------------


   subtype Scheduling_Sequence is A_Time_Unit.Indexed_Table;
   type Scheduling_Sequence_Ptr is access Scheduling_Sequence; 


   procedure Free is 
   new Unchecked_Deallocation
      (Scheduling_Sequence, Scheduling_Sequence_Ptr);



   --------------------------------------------------------
   --------------------------------------------------------
   -- Services to do analysis of scheduling sequences
   --------------------------------------------------------
   --------------------------------------------------------



   --------------------------------------------------------
   -- Processor analysis
   --------------------------------------------------------


   -- Compute number of task preemption
   --
   function Number_Of_Preemption_From_Simulation (
         Sched    : in     Scheduling_Sequence_Ptr; 
         My_Tasks : in     Tasks_Set                ) 
     return Natural; 


   -- Compute number of switch context
   --
   function Number_Of_Switch_Context_From_Simulation (
         Sched : in     Scheduling_Sequence_Ptr ) 
     return Natural; 




   --------------------------------------------------------
   -- Buffer analysis (buffer utilization factor)
   --------------------------------------------------------


   type Buffer_Size_Item is 
      record 
         Time : Natural;  
         Size : Natural;  
      end record; 

   procedure Initialize (
         B :    out Buffer_Size_Item ); 
   procedure Put (
         B : in     Buffer_Size_Item ); 




   package Buffer_Size_Package is new Tables(Buffer_Size_Item,
      Config.Max_Scheduling_Period,  Put);
   use Buffer_Size_Package;

   subtype Buffer_Size_Table is Buffer_Size_Package.Table;
   subtype Buffer_Size_Range is Buffer_Size_Package.Table_Range;
   type Buffer_Size_Table_Ptr is access Buffer_Size_Table; 

   procedure Initialize (
         Wt :    out Double ); 
   procedure Put (
         Wt : in     Double ); 

   package Waiting_Time_Package is new Tables(Double,
      Config.Max_Scheduling_Period,  Put);
   use Waiting_Time_Package;

   subtype Waiting_Time_Table is Waiting_Time_Package.Table;
   subtype Waiting_Time_Range is Waiting_Time_Package.Table_Range;
   type Waiting_Time_Table_Ptr is access Waiting_Time_Table; 


   -- Compute from a scheduling sequence buffer size for each
   -- time unit
   --
   procedure Compute_Buffer_Size_From_Simulation (
         Sched   : in     Scheduling_Sequence_Ptr; 
         My_Buff : in     Buffer_Ptr;              
         Buff    : in out Buffer_Size_Table_Ptr    ); 


   -- Compute, from a buffer_Size_Table, maximum and average
   -- performance criterion
   --
   function Compute_Average_Buffer_Size_From_Simulation (
         Buff : in     Buffer_Size_Table_Ptr ) 
     return Double; 

   function Compute_Maximum_Buffer_Size_From_Simulation (
         Buff : in     Buffer_Size_Table_Ptr ) 
     return Natural; 

   function Compute_Maximum_Waiting_Time_From_Simulation (
         Buff                    : in     Buffer_Size_Table_Ptr; 
         Average_Consumer_Period : in     Double                 ) 
     return Natural; 

   function Compute_Average_Waiting_Time_From_Simulation (
         Buff                    : in     Buffer_Size_Table_Ptr; 
         Average_Consumer_Period : in     Double                 ) 
     return Double; 





   procedure Free is 
   new Unchecked_Deallocation
      (Buffer_Size_Table, Buffer_Size_Table_Ptr);

   procedure Free is 
   new Unchecked_Deallocation
      (Waiting_Time_Table, Waiting_Time_Table_Ptr);

   No_Performance_Criteria : exception;  

   -- copy consumer response time from simulation of a scheduling
   -- in cheddar to a qs (example : MP1 criteria computation)
   --
   procedure Consumer_Response_Time (
         Sched              : in     Scheduling_Sequence_Ptr;      
         Consumer_Resp_Time : in out Resp_Time_Consumer_Table_Ptr; 
         My_Consumer_Task   : in     Task_Ptr                      ); 

   package Generic_Index_Lst is new Lists (
      Put     => Put,             
      Element => Double_Ptr,      
      Free    => Double_Util.Free);

   use Generic_Index_Lst;



   --------------------------------------------------------
   -- Task analysis (eg. response time from simulation)
   --------------------------------------------------------


   -- Compute bound on response time from simulation
   --
   procedure Response_Time_By_Simulation (
         My_Task :        Task_Ptr;                
         Sched   : in     Scheduling_Sequence_Ptr; 
         Msg     : in out Unbounded_String;        
         Max     : in out Natural;                 
         Min     : in out Natural;                 
         Average : in out Double                   ); 


   -- Compute probability distribution of response time
   --
   type Density_Item is 
      record 
         Response_Time : Natural;  
         Probability   : Double;  
      end record; 


   procedure Initialize (
         D : in out Density_Item ); 
   procedure Put (
         D : in     Density_Item ); 


   package Density_Table_Package is new Tables(
      Density_Item, Config.Max_Density_Size, Put);
   use Density_Table_Package;
   subtype Density_Table is Density_Table_Package.Table;
   subtype Density_Range is Density_Table_Package.Table_Range;


   package Densities_Table_Package is new Indexed_Tables(Density_Table,
      Task_Ptr, Config.Max_Tasks, 0, Put, Initialize, Put_Name);
   use  Densities_Table_Package;
   subtype Densities_Table is Densities_Table_Package.Indexed_Table;
   subtype Densities_Range is
      Densities_Table_Package.Indexed_Table_Range;


   procedure Compute_Response_Time_Distribution (
         My_Tasks       : in     Tasks_Set;               
         Sched          : in     Scheduling_Sequence_Ptr; 
         Processor_Name :        Unbounded_String;        
         Result         : in out Densities_Table          ); 



   -- Compute the response time for each task activation
   --
   package Task_Occurence_Table_Package is new Tables(
      Task_Occurence, Config.Max_Scheduling_Period, Put);
   use Task_Occurence_Table_Package;

   subtype Task_Occurence_Table is Task_Occurence_Table_Package.Table;
   subtype Task_Occurence_Range is Task_Occurence_Table_Package.
      Table_Range;
   type Task_Occurence_Table_Ptr is access Task_Occurence_Table; 



   procedure Free is 
   new Unchecked_Deallocation
      (Task_Occurence_Table, Task_Occurence_Table_Ptr);


   procedure All_Response_Time_By_Simulation (
         My_Task : in     Task_Ptr;                
         Sched   : in     Scheduling_Sequence_Ptr; 
         Result  : in out Task_Occurence_Table_Ptr ); 





   --------------------------------------------------------
   -- Shared resource  analysis (eg. blocking time from simulation)
   --------------------------------------------------------







   -- Compute blocking time from scheduling simulaion
   --
   package Double_Tasks_Parameters is new Indexed_Tables(Double,
      Task_Ptr,
      Config.Max_Tasks, 0, Double_Util.Put, Double_Util.Initialize,
      Put_Name);

   subtype Blocking_Time_Table is Double_Tasks_Parameters.Indexed_Table;
   subtype Blocking_Time_Range is Double_Tasks_Parameters.
      Indexed_Table_Range;


   procedure Blocking_Time_From_Simulation (
         My_Task      : in     Task_Ptr;                
         My_Resources : in     Resources_Set;           
         Sched        : in     Scheduling_Sequence_Ptr; 
         Max          : in out Natural;                 
         Min          : in out Natural;                 
         Average      : in out Double                   ); 



   -- Check if deadlocks on shared resources exist
   --
   type Task_Name_Table is array (1 .. Config.Max_Tasks) of Unbounded_String; 
   type Resource_Name_Table is array (1 .. Config.Max_Resources) of Unbounded_String; 

   type Deadlock_Item is 
      record 
         Time          : Natural;  
         Task_Name     : Task_Name_Table;  
         Resource_Name : Resource_Name_Table;  
      end record; 

   type Deadlock_Item_Ptr is access all Deadlock_Item; 
   procedure Free is 
   new Unchecked_Deallocation
      (Deadlock_Item, Deadlock_Item_Ptr );

   procedure Initialize (
         D : in out Deadlock_Item ); 
   procedure Put (
         D : in     Deadlock_Item ); 
   procedure Put (
         D : in     Deadlock_Item_Ptr ); 


   package Deadlock_List_Package is new Lists(
      Deadlock_Item_Ptr, Put, Free);
   use Deadlock_List_Package;
   subtype Deadlock_List is Deadlock_List_Package.List;



   procedure Deadlock_From_Simulation (
         My_Tasks       : in     Tasks_Set;               
         My_Resources   : in     Resources_Set;           
         Processor_Name : in     Unbounded_String;        
         Sched          : in     Scheduling_Sequence_Ptr; 
         Result         : in out Deadlock_List            ); 





   -- Check if priority inversions exists in a scheduling simulation
   --

   type Priority_Inversion_Item is 
      record 
         Start_Time    : Natural;  
         End_Time      : Natural;  
         Task_Name     : Unbounded_String;  
         Resource_Name : Unbounded_String;  
      end record; 

   type Priority_Inversion_Item_Ptr is access all Priority_Inversion_Item; 
   procedure Free is 
   new Unchecked_Deallocation (Priority_Inversion_Item,
      Priority_Inversion_Item_Ptr);


   procedure Initialize (
         D : in out Priority_Inversion_Item ); 
   procedure Put (
         D : in     Priority_Inversion_Item ); 
   procedure Put (
         D : in     Priority_Inversion_Item_Ptr ); 


   package Priority_Inversion_List_Package is new Lists(
      Priority_Inversion_Item_Ptr, Put, Free);
   use Priority_Inversion_List_Package;
   subtype Priority_Inversion_List is
      Priority_Inversion_List_Package.List;


   function Priority_Inversion_From_Simulation (
         My_Tasks       : in     Tasks_Set;              
         My_Resources   : in     Resources_Set;          
         Processor_Name : in     Unbounded_String;       
         Sched          : in     Scheduling_Sequence_Ptr ) 
     return Priority_Inversion_List; 




end Scheduling_Sequences;
